# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import functools
from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import check_file_or_directory_path
from msprobe.core.common.file_utils import save_npy


class FrameworkDescriptor:
    def __get__(self, instance, owner):
        if owner._framework is None:
            owner.import_framework()
        return owner._framework


class FmkAdp:
    fmk = Const.PT_FRAMEWORK
    supported_fmk = [Const.PT_FRAMEWORK, Const.MS_FRAMEWORK]
    supported_dtype_list = ["bfloat16", "float16", "float32", "float64"]
    _framework = None
    framework = FrameworkDescriptor()

    @classmethod
    def import_framework(cls):
        if cls.fmk == Const.PT_FRAMEWORK:
            import torch
            cls._framework = torch
        elif cls.fmk == Const.MS_FRAMEWORK:
            import mindspore
            cls._framework = mindspore
        else:
            raise Exception(f"init framework adapter error, not in {cls.supported_fmk}")

    @classmethod
    def set_fmk(cls, fmk=Const.PT_FRAMEWORK):
        if fmk not in cls.supported_fmk:
            raise Exception(f"init framework adapter error, not in {cls.supported_fmk}")
        cls.fmk = fmk
        cls._framework = None  # 重置框架，以便下次访问时重新导入

    @classmethod
    def get_rank(cls):
        if cls.fmk == Const.PT_FRAMEWORK:
            return cls.framework.distributed.get_rank()
        return cls.framework.communication.get_rank()

    @classmethod
    def get_rank_id(cls):
        if cls.is_initialized():
            return cls.get_rank()
        return 0

    @classmethod
    def is_initialized(cls):
        if cls.fmk == Const.PT_FRAMEWORK:
            return cls.framework.distributed.is_initialized()
        return cls.framework.communication.GlobalComm.INITED

    @classmethod
    def is_nn_module(cls, module):
        if cls.fmk == Const.PT_FRAMEWORK:
            return isinstance(module, cls.framework.nn.Module)
        return isinstance(module, cls.framework.nn.Cell)

    @classmethod
    def is_tensor(cls, tensor):
        if cls.fmk == Const.PT_FRAMEWORK:
            return isinstance(tensor, cls.framework.Tensor)
        return isinstance(tensor, cls.framework.Tensor)

    @classmethod
    def process_tensor(cls, tensor, func):
        if cls.fmk == Const.PT_FRAMEWORK:
            if not tensor.is_floating_point() or tensor.dtype == cls.framework.float64:
                tensor = tensor.float()
            return float(func(tensor))
        return float(func(tensor).asnumpy())

    @classmethod
    def tensor_max(cls, tensor):
        return cls.process_tensor(tensor, lambda x: x.max())

    @classmethod
    def tensor_min(cls, tensor):
        return cls.process_tensor(tensor, lambda x: x.min())

    @classmethod
    def tensor_mean(cls, tensor):
        return cls.process_tensor(tensor, lambda x: x.mean())

    @classmethod
    def tensor_norm(cls, tensor):
        return cls.process_tensor(tensor, lambda x: x.norm())
    
    @classmethod
    def save_tensor(cls, tensor, filepath):
        if cls.fmk == Const.PT_FRAMEWORK:
            tensor_npy = tensor.cpu().detach().float().numpy()
        else:
            tensor_npy = tensor.asnumpy()
        save_npy(tensor_npy, filepath)

    @classmethod
    def dtype(cls, dtype_str):
        if dtype_str not in cls.supported_dtype_list:
            raise Exception(f"{dtype_str} is not supported by adapter, not in {cls.supported_dtype_list}")
        return getattr(cls.framework, dtype_str)

    @classmethod
    def named_parameters(cls, module):
        if cls.fmk == Const.PT_FRAMEWORK:
            if not isinstance(module, cls.framework.nn.Module):
                raise Exception(f"{module} is not a torch.nn.Module")
            return module.named_parameters()
        if not isinstance(module, cls.framework.nn.Cell):
            raise Exception(f"{module} is not a mindspore.nn.Cell")
        return module.parameters_and_names()

    @classmethod
    def iter_named_modules(cls, model):
        """
        Unified module iterator across frameworks.

        Returns an iterable of (name, module) pairs using framework-specific
        traversal APIs:
        - PyTorch: model.named_modules()
        - MindSpore: model.cells_and_names()
        """
        if cls.fmk == Const.PT_FRAMEWORK:
            if not hasattr(model, "named_modules"):
                raise Exception(f"{model} does not implement named_modules()")
            return model.named_modules()
        # MindSpore
        if not hasattr(model, "cells_and_names"):
            raise Exception(f"{model} does not implement cells_and_names()")
        return model.cells_and_names()

    @classmethod
    def register_forward_pre_hook(cls, module, hook, with_kwargs=False):
        if cls.fmk == Const.PT_FRAMEWORK:
            if not isinstance(module, cls.framework.nn.Module):
                raise Exception(f"{module} is not a torch.nn.Module")
            module.register_forward_pre_hook(hook, with_kwargs=with_kwargs)
        else:
            if not isinstance(module, cls.framework.nn.Cell):
                raise Exception(f"{module} is not a mindspore.nn.Cell")
            original_construct = module.construct

            @functools.wraps(original_construct)
            def new_construct(*args, **kwargs):
                if with_kwargs:
                    hook(module, args, kwargs)
                else:
                    hook(module, args)
                return original_construct(*args, **kwargs)

            module.construct = new_construct

    @classmethod
    def load_checkpoint(cls, path, to_cpu=True, weights_only=True):
        check_file_or_directory_path(path, is_strict=not weights_only)
        if cls.fmk == Const.PT_FRAMEWORK:
            try:
                if to_cpu:
                    return cls.framework.load(path, map_location=cls.framework.device("cpu"), weights_only=weights_only)
                else:
                    return cls.framework.load(path, weights_only=weights_only)
            except Exception as e:
                raise RuntimeError(f"load pt file {path} failed: {e}") from e
        return mindspore.load_checkpoint(path)
    
    @classmethod
    def asnumpy(cls, tensor):
        if cls.fmk == Const.PT_FRAMEWORK:
            return tensor.float().numpy()
        return tensor.float().asnumpy()


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.log import logger
from msprobe.mindspore.common.const import Const
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig
from msprobe.mindspore.dump.exception_dump.kernel_graph_exception_dump import KernelGraphExceptionDump


class ExceptionDumpToolFactory:
    tools = {
        Const.CELL: {
            Const.GRAPH_KBYK_MODE: None,
            Const.GRAPH_GE_MODE: None,
            Const.PYNATIVE_MODE: None
        },
        Const.API: {
            Const.GRAPH_KBYK_MODE: None,
            Const.GRAPH_GE_MODE: None,
            Const.PYNATIVE_MODE: None
        },
        Const.KERNEL: {
            Const.GRAPH_KBYK_MODE: KernelGraphExceptionDump,
            Const.GRAPH_GE_MODE: None,
            Const.PYNATIVE_MODE: KernelGraphExceptionDump
        }
    }

    @staticmethod
    def create(config: DebuggerConfig):
        tool = ExceptionDumpToolFactory.tools.get(config.level)
        if not tool:
            raise Exception("Valid level is needed.")
        tool = tool.get(config.execution_mode)
        if not tool:
            logger.error(f"Exception dump is not supported in {config.execution_mode} mode "
                         f"when level is {config.level}.")
            raise ValueError
        return (tool(config),)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import json
from pathlib import Path
from werkzeug import wrappers, Response, exceptions
from ..utils.request_method import request_method
from ..service import ServiceFactory, GraphServiceStrategy
from ..utils.file_check_wrapper import check_file_type
from ..utils.graph_utils import GraphUtils
from ..utils.constant import DataType
from ..utils.global_state import GraphState
from ..utils.constant import security_headers


class GraphView:
    service_factory = ServiceFactory()

    # 静态文件路由
    @staticmethod
    @wrappers.Request.application
    def static_file_route(request):
        filename = os.path.basename(request.path)

        extension = os.path.splitext(filename)[1]
        if extension == '.html':
            content_type = 'text/html'
        elif extension == '.js':
            content_type = 'application/javascript'
        else:
            content_type = 'application/octet-stream'

        try:
            # 添加白名单校验
            if filename != 'index.html' and filename != 'index.js':
                raise exceptions.NotFound('404 Not Found') from e
            current_dir = Path(__file__).resolve().parent
            server_dir = current_dir.parent.parent
            dir_path = (server_dir / "static").resolve()
            filepath = (server_dir / "static" / filename).resolve()  # resolve() 规范化路径
            if not GraphUtils.is_relative_to(filepath, dir_path):
                raise exceptions.NotFound('404 Not Found')
            # 前端打包后产生的内部输入文件，不需要要安全校验
            with open(filepath, 'rb') as infile:
                contents = infile.read()
        except IOError as e:
            raise exceptions.NotFound('404 Not Found') from e
        return Response(contents, content_type=content_type, headers=security_headers)

    @staticmethod
    @wrappers.Request.application
    @request_method('GET')
    def load_meta_dir(request):
        """Scan logdir for directories containing .vis files, modified to return a tuple of (run, tag)."""
        result = GraphServiceStrategy.load_meta_dir()
        return Response(json.dumps(result), content_type="application/json", headers=security_headers)

    # 读取当前图数据
    @staticmethod
    @wrappers.Request.application
    @request_method('GET')
    def load_graph_data(request):
        meta_data = {
            'run': request.args.get('run'),
            'tag': request.args.get('tag'),
            'type': request.args.get('type'),
        }
        lang = request.args.get('lang')
        GraphState.set_global_value('lang', lang)
        strategy = GraphView._get_strategy(meta_data)
        if meta_data.get('type') == DataType.DB.value:
            result = strategy.load_graph_data()
            return Response(result, content_type="application/json", headers=security_headers)
        elif meta_data.get('type') == DataType.JSON.value:
            return Response(
                strategy.load_graph_data(),
                mimetype='text/event-stream',
                headers={
                    'Cache-Control': 'no-cache',
                    'Connection': 'close',  # TCP链接不复用，请求结束释放资源
                } | security_headers
            )
        else: 
            result = {'success': False, 'error': GraphUtils.t('typeError')}
            return Response(json.dumps(result), content_type="application/json", headers=security_headers)

    # 获取当前图数据配置信息
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def load_graph_config_info(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        meta_data = data.get('metaData')
        strategy = GraphView._get_strategy(meta_data)
        result = strategy.load_graph_config_info()
        # 创建响应对象
        response = Response(json.dumps(result), content_type="application/json", headers=security_headers)
        return response

    # 获取当前图所有节点列表
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def load_graph_all_node_list(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        meta_data = data.get('metaData')
        strategy = GraphView._get_strategy(meta_data)
        result = strategy.load_graph_all_node_list(meta_data)
        response = Response(json.dumps(result), content_type="application/json", headers=security_headers)
        return response          

    # 根据精度误差搜索节点
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def search_node(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'))
        meta_data = data.get("metaData")
        search_type = data.get("type")
        values = data.get("values")
        strategy = GraphView._get_strategy(meta_data)
        if search_type == 'precision':
            result = strategy.search_node_by_precision(meta_data, values)
        elif search_type == 'overflow':
            result = strategy.search_node_by_overflow(meta_data, values)
        else:
            result = {
                'success': False,
                'error': GraphUtils.t('searchTypeError')
            }
        return Response(json.dumps(result), content_type="application/json", headers=security_headers) 

    # 更新误差节点
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def update_precision_error(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'))
        meta_data = data.get('metaData')
        filter_value = data.get("filterValue")
        strategy = GraphView._get_strategy(meta_data)
        result = strategy.update_precision_error(meta_data, filter_value)
        return Response(json.dumps(result), content_type="application/json", headers=security_headers)

    # 展开关闭节点
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def change_node_expand_state(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        node_info = GraphUtils.safe_get_node_info(data)
        result = {'success': False, 'error': ''}
        meta_data = data.get('metaData')
        if node_info is None or not isinstance(node_info, dict):
            result['error'] = GraphUtils.t('nodeInfoError') 
        strategy = GraphView._get_strategy(meta_data)
        hierarchy = strategy.change_node_expand_state(node_info, meta_data)
        return Response(json.dumps(hierarchy), content_type="application/json", headers=security_headers)

    # 更新当前图节点信息
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def update_hierarchy_data(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        graph_type = data.get("graphType")
        meta_data = data.get('metaData')
        strategy = GraphView._get_strategy(meta_data)
        hierarchy = strategy.update_hierarchy_data(graph_type)
        return Response(json.dumps(hierarchy), content_type="application/json", headers=security_headers)

    # 获取当前节点对应节点的信息看板数据
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def get_node_info(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        node_info = GraphUtils.safe_get_node_info(data)
        result = {'success': False, 'error': ''}
        meta_data = data.get('metaData')
        if node_info is None or not isinstance(node_info, dict):
            result['error'] = GraphUtils.t('nodeInfoError') 
            return Response(result, content_type="application/json", headers=security_headers)
        strategy = GraphView._get_strategy(meta_data)
        node_detail = strategy.get_node_info(node_info, meta_data)
        return Response(json.dumps(node_detail), content_type="application/json", headers=security_headers)

    # 根据配置文件添加匹配节点
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def add_match_nodes_by_config(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        config_file = data.get('configFile')
        meta_data = data.get('metaData')
        strategy = GraphView._get_strategy(meta_data)
        match_result = strategy.add_match_nodes_by_config(config_file, meta_data)
        return Response(json.dumps(match_result), content_type="application/json", headers=security_headers)

    # 添加匹配节点
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def add_match_nodes(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        npu_node_name = data.get("npuNodeName")
        bench_node_name = data.get("benchNodeName")
        meta_data = data.get('metaData')
        is_match_children = data.get("isMatchChildren")
        strategy = GraphView._get_strategy(meta_data)
        match_result = strategy.add_match_nodes(npu_node_name, bench_node_name, meta_data, is_match_children)
        return Response(json.dumps(match_result), content_type="application/json", headers=security_headers)

    # 取消节点匹配
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def delete_match_nodes(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        npu_node_name = data.get("npuNodeName")
        bench_node_name = data.get("benchNodeName")
        meta_data = data.get('metaData')
        is_unmatch_children = data.get("isUnMatchChildren")
        strategy = GraphView._get_strategy(meta_data)
        match_result = strategy.delete_match_nodes(npu_node_name, bench_node_name, meta_data, is_unmatch_children)
        return Response(json.dumps(match_result), content_type="application/json", headers=security_headers)

    # 保存匹配节点列表
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def save_data(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        meta_data = data.get('metaData')
        strategy = GraphView._get_strategy(meta_data)
        save_result = strategy.save_data(meta_data)
        return Response(json.dumps(save_result), content_type="application/json", headers=security_headers)

    # 更新颜色信息
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def update_colors(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        meta_data = data.get('metaData')
        colors = GraphUtils.safe_json_loads(data.get('colors'))
        success, error_msg, colors = GraphUtils.validate_colors_param(colors)
        if not success:
            result = {
                'success': False,
                'error': error_msg
            }
            return Response(json.dumps(result), content_type="application/json", headers=security_headers)
        strategy = GraphView._get_strategy(meta_data, no_tag=True)
        update_result = strategy.update_colors(colors)
        return Response(json.dumps(update_result), content_type="application/json", headers=security_headers)

    # 保存匹配关系
    @staticmethod
    @wrappers.Request.application
    @check_file_type
    @request_method('POST')
    def save_matched_relations(request):
        data = GraphUtils.safe_json_loads(request.get_data().decode('utf-8'), {})
        meta_data = data.get('metaData')
        strategy = GraphView._get_strategy(meta_data)
        save_result = strategy.save_matched_relations(meta_data)
        return Response(json.dumps(save_result), content_type="application/json", headers=security_headers)

    @staticmethod
    def _get_strategy(meta_data, no_tag=False):
        run = meta_data.get('run')
        data_type = meta_data.get('type')
        if no_tag:
            return GraphView.service_factory.create_strategy_without_tag(data_type, run)
        
        tag = meta_data.get('tag')
        return GraphView.service_factory.create_strategy(data_type, run, tag)



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os


CANN_PATH = os.environ.get('ASCEND_TOOLKIT_HOME', "/usr/local/Ascend/ascend-toolkit/latest")


class CmpArgsAdapter:
    def __init__(self,
                 golden_path,
                 target_path,
                 input_data="",
                 cann_path=CANN_PATH,
                 output_path="./output",
                 input_shape="",
                 rank="0",
                 output_size="",
                 dym_shape_range="",
                 onnx_fusion_switch=True
                 ):
        self.golden_path = golden_path
        self.target_path = target_path
        self.input_data = input_data
        self.cann_path = cann_path
        self.output_path = output_path
        self.input_shape = input_shape
        self.rank = rank
        self.output_size = output_size
        self.dym_shape_range = dym_shape_range
        self.onnx_fusion_switch = onnx_fusion_switch



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re

from msprobe.core.compare.acc_compare import ModeConfig
from msprobe.core.compare.multiprocessing_compute import CompareRealData
from msprobe.core.compare.utils import read_op, merge_tensor, get_accuracy, make_result_table
from msprobe.core.common.utils import set_dump_path, get_dump_mode
from msprobe.visualization.utils import GraphConst
from msprobe.core.common.const import Const
from msprobe.core.compare.indicator_analysis.calculator import calculate_result

# 用于将节点名字解析成对应的NodeOp的规则
op_patterns = [
    # NodeOp.module
    r'^(Module.|Cell.|optimizer|clip_grad|DefaultModel)',
    # NodeOp.function_api
    r'^(Tensor.|Torch.|Functional.|NPU.|VF.|Distributed.|Aten.|Mint.|Primitive.|Jit.|MintFunctional.|MindSpeed.)'
]


def get_compare_mode(dump_path_param):
    """
    获得比较模式，包括summary、MD5和真实数据三种模式
    Args:
        dump_path_param: 调用acc_compare接口所依赖的参数
    Returns: 0 summary mode, 1 md5 mode, 2 true data mode
    """
    set_dump_path(dump_path_param)
    dump_mode = get_dump_mode(dump_path_param)
    compare_mode = GraphConst.DUMP_MODE_TO_GRAPHCOMPARE_MODE_MAPPING.get(dump_mode)
    return compare_mode


def run_real_data(dump_path_param, csv_path, framework, is_cross_frame=False):
    """
    多进程运行生成真实数据
    Args:
        dump_path_param: 调用acc_compare接口所依赖的参数
        csv_path: 生成文件路径
        framework: 框架类型, pytorch或mindspore
        is_cross_frame: 是否进行跨框架比对，仅支持mindspore比pytorch, 其中pytorch为标杆
    """
    config_dict = {
        'stack_mode': False,
        'auto_analyze': True,
        'fuzzy_match': False,
        'dump_mode': Const.ALL
    }
    mode_config = ModeConfig(**config_dict)

    if framework == Const.PT_FRAMEWORK:
        from msprobe.pytorch.compare.pt_compare import read_real_data
        return CompareRealData(read_real_data, mode_config, is_cross_frame).do_multi_process(dump_path_param, csv_path)
    else:
        from msprobe.mindspore.compare.ms_compare import read_real_data
        return CompareRealData(read_real_data, mode_config, is_cross_frame).do_multi_process(dump_path_param, csv_path)


def get_input_output(node_data, node_id):
    """
    将dump的原始数据进行拆解，分解为output和input两个数据
    Args:
        node_data: 属于单个节点的dump数据
        node_id: 节点名字
    """
    input_data = {}
    output_data = {}
    op_parsed_list = read_op(node_data, node_id)
    for item in op_parsed_list:
        full_op_name = item.get('full_op_name', '')
        if not full_op_name:
            continue
        if GraphConst.OUTPUT in full_op_name and GraphConst.INPUT not in full_op_name:
            output_data[full_op_name] = item
        else:
            name = item.get('data_name')
            # 节点参数名称尽量使用落盘数据的名称
            if isinstance(name, str) and name != '-1':
                input_data[name.rsplit(Const.SEP, 1)[0]] = item
            else:
                input_data[full_op_name] = item
    return input_data, output_data


def format_node_data(data_dict, node_id=None, compare_mode=None):
    """
    删除节点数据中不需要展示的字段
    """
    del_list = ['state', 'full_op_name']
    if GraphConst.MD5_COMPARE != compare_mode:
        del_list.append(Const.MD5)
    if node_id and GraphConst.BATCH_P2P in node_id:
        del_list.extend(['op', 'peer', 'tag', 'group_id'])
    for _, value in data_dict.items():
        if not isinstance(value, dict):
            continue
        for item in del_list:
            if item in value:
                del value[item]
        _format_data(value)
    return data_dict


def compare_node(node_n, node_b, compare_mode):
    """
    调用acc_compare.py中的get_accuracy获得精度对比指标
    真实数据对比模式无法获得精度对比指标，需要调用多进程比对接口
    Returns: 包含参数信息和对比指标（真实数据对比模式除外）的list
    """
    dump_mode = GraphConst.GRAPHCOMPARE_MODE_TO_DUMP_MODE_TO_MAPPING.get(compare_mode)
    merge_n = _parse_node(node_n, dump_mode)
    merge_b = _parse_node(node_b, dump_mode)
    result = []
    get_accuracy(result, merge_n, merge_b, dump_mode)
    return result


def get_api_indicator_info(compare_mode, result, parallel_merge=False):
    """
    得到一个api或模块的指标和异常信息
    """
    dump_mode = GraphConst.GRAPHCOMPARE_MODE_TO_DUMP_MODE_TO_MAPPING.get(compare_mode)
    return calculate_result(result, dump_mode, parallel_merge)


def get_real_api_data_list(node, compare_data_dict: dict):
    """
    真实数据模式下，得到的比对结果包含了所有api参数，需要转换为其中某一个api参数列表
    """
    api_data_list = []
    keys = list(node.input_data.keys()) + list(node.output_data.keys())
    for key in keys:
        if key in compare_data_dict:
            api_data_list.append(compare_data_dict.get(key))
    return api_data_list


def _parse_node(node, dump_mode):
    """
    转换节点，使其能够作为acc_compare.py中的get_accuracy的入参
    """
    op_parsed_list = []
    op_parsed_list.extend(node.input_data.values())
    op_parsed_list.extend(node.output_data.values())
    result = merge_tensor(op_parsed_list, dump_mode)
    if not result:
        result['op_name'] = []
    return result


def _format_decimal_string(s):
    """
    使用正则表达式匹配包含数字、小数点和可选的百分号的字符串
    """
    pattern = re.compile(r'^\d{1,20}\.\d{1,20}%?$')
    matches = pattern.findall(s)
    for match in matches:
        is_percent = match.endswith('%')
        number_str = match.rstrip('%')
        decimal_part = number_str.split('.')[1]
        # 如果小数位数大于6，进行处理
        if len(decimal_part) > GraphConst.ROUND_TH:
            number_float = float(number_str)
            formatted_number = f"{number_float:.{GraphConst.ROUND_TH}f}"
            # 如果原来是百分数，加回百分号
            if is_percent:
                formatted_number += '%'
            # 替换原字符串中的数值部分
            s = s.replace(match, formatted_number)
    return s


def _format_data(data_dict):
    """
    格式化数据，小数保留6位，处理一些异常值
    """
    pattern = r'^[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)$'
    all_null = False

    keys_to_keep = ['type', 'group_ranks', 'group_id', 'data_name']
    if data_dict.get('type') == 'torch.ProcessGroup':
        keys_to_remove = [key for key in data_dict if key not in keys_to_keep]
        for key in keys_to_remove:
            del data_dict[key]

    for key, value in data_dict.items():
        if isinstance(value, str):
            # 将单引号删掉，None换成null避免前端解析错误
            value = value.replace("'", "").replace(GraphConst.NONE, GraphConst.NULL)
            value = _format_decimal_string(value)
        elif value is None or value == ' ':
            value = GraphConst.NULL
        # 科学计数法1.123123123123e-11，格式化为1.123123e-11
        elif isinstance(value, float) and len(str(value)) < GraphConst.STR_MAX_LEN and re.match(pattern, str(value)):
            value = "{:.6e}".format(value)
        elif isinstance(value, float):
            value = round(value, GraphConst.ROUND_TH)
        # Inf会走入这里，确保转成Inf。另外给其他不符合预期的类型做兜底方案
        if key != GraphConst.ERROR_KEY:
            # 除了error_key不转str，其他都转str, 避免前端解析错误
            value = str(value)
        # max为null, 意味着这个参数值为null
        if key == Const.MAX and value == GraphConst.NULL:
            all_null = True
        data_dict[key] = value
    # 字典里的value全null，只保留一个null
    if all_null:
        data_dict.clear()
        data_dict[GraphConst.VALUE] = GraphConst.NULL


def get_csv_df(stack_mode, csv_data, compare_mode):
    """
    调用acc接口写入csv
    """

    dump_mode = GraphConst.GRAPHCOMPARE_MODE_TO_DUMP_MODE_TO_MAPPING.get(compare_mode)
    return make_result_table(csv_data, dump_mode, stack_mode)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from mindspore import dtype as mstype

from msprobe.core.common.const import Const as CoreConst


class Const:
    CELL = "cell"
    API = "api"
    KERNEL = "kernel"
    CELL_AND_API = 'cell_and_api'
    TOOL_LEVEL_DICT = {
        CoreConst.LEVEL_L0: CELL,
        CoreConst.LEVEL_L1: API,
        CoreConst.LEVEL_L2: KERNEL,
        CoreConst.LEVEL_MIX: CELL_AND_API
    }

    PYNATIVE_MODE = CoreConst.PYNATIVE_MODE
    GRAPH_MODE = "graph"
    GRAPH_GE_MODE = "graph_ge"
    GRAPH_KBYK_MODE = "graph_kbyk"
    PYNATIVE_GRAPH_MODE = CoreConst.PYNATIVE_GRAPH_MODE

    JIT_LEVEL = "jit_level"
    JIT_LEVEL_O0 = "O0"
    JIT_LEVEL_O1 = "O1"
    JIT_LEVEL_O2 = "O2"
    ASCEND_910A = "ascend910"

    OPS_PREFIX = "mindspore.ops."
    TENSOR_PREFIX = "mindspore.Tensor."
    MINT_PREFIX = "mindspore.mint."
    MINT_NN_FUNC_PREFIX = "mindspore.mint.nn.functional."

    TENSOR_DATA_PREFIX = "Tensor."
    STUB_TENSOR_DATA_PREFIX = "Tensor."
    OPS_DATA_PREFIX = "Functional."
    MINT_DATA_PREFIX = "Mint."
    MINT_NN_FUNC_DATA_PREFIX = "MintFunctional."
    DISTRIBUTED_DATA_PREFIX = "Distributed."
    TORCH_DATA_PREFIX = "Torch."
    TORCH_NPU_DATA_PREFIX = "NPU."

    SUPPORTED_API_LIST_FILE = "support_wrap_ops.yaml"
    SUPPORTED_TENSOR_LIST_KEY = "tensor"
    SUPPORTED_OPS_LIST_KEY = "ops"
    SUPPORTED_MINT_LIST_KEY = "mint.ops"
    SUPPORTED__MINT_NN_FUNC_LIST_KEY = "mint.nn.functional"
    SUPPORTED_COMM_LIST_KEY = "communication.comm_func"

    DROPOUT_API_NAME_PREFIX = "dropout"

    GRAPH_DATA_MODE_LIST = [CoreConst.ALL, CoreConst.INPUT, CoreConst.OUTPUT]
    GRAPH_CELL_DUMP_DATA_MODE_LIST = [CoreConst.ALL, CoreConst.FORWARD, CoreConst.BACKWARD]

    HOOK_MS_PREFIX_DICT = {
        OPS_DATA_PREFIX: OPS_PREFIX,
        TENSOR_DATA_PREFIX: TENSOR_PREFIX,
        MINT_DATA_PREFIX: MINT_PREFIX,
        MINT_NN_FUNC_DATA_PREFIX: MINT_NN_FUNC_PREFIX
    }

    NonDifferentiableType = (
        mstype.bool_, mstype.int8, mstype.byte, mstype.uint8, mstype.ubyte,
        mstype.int16, mstype.short, mstype.uint16, mstype.ushort,
        mstype.int32, mstype.intc, mstype.uint32, mstype.uintc,
        mstype.int64, mstype.intp, mstype.uint64, mstype.uintp
    )

    FP8_TYPE_LIST = ["Float8E5M2", "Float8E4M3FN", "HiFloat8"]


class MsCompareConst:
    # api_info field
    MINT = "Mint"
    MINT_FUNCTIONAL = "MintFunctional"
    TENSOR_API = "Tensor"
    FUNCTIONAL_API = "Functional"
    FUSION_API = "FUSION"

    API_NAME_STR_LENGTH = 4
    MAX_RECURSION_DEPTH = 20

    # Mindtorch api_info field
    MINDTORCH_TENSOR = "Tensor"
    MINDTORCH = "Torch"
    MINDTORCH_FUNC = "Functional"
    MINDTORCH_NPU = "NPU"
    MINDTORCH_DIST = "Distributed"

    MT_VALID_API_TYPES = [
        MINDTORCH, MINDTORCH_FUNC, MINDTORCH_TENSOR
    ]
    SUPPORTED_FUSION_LIST = ["flash_attention_score"]

    TASK_FIELD = "task"
    STATISTICS_TASK = "statistics"
    FRAMEWORK = "framework"
    TENSOR_TASK = "tensor"
    DUMP_DATA_DIR_FIELD = "dump_data_dir"
    DATA_FIELD = "data"

    # supported api yaml
    SUPPORTED_API_LIST_FILE = "checker_support_api.yaml"
    SUPPORTED_TENSOR_LIST_KEY = "tensor"

    # detail_csv
    DETAIL_CSV_API_NAME = "API Name"
    DETAIL_CSV_BENCH_DTYPE = "Bench Dtype"
    DETAIL_CSV_TESTED_DTYPE = "Tested Dtype"
    DETAIL_CSV_SHAPE = "Shape"
    DETAIL_CSV_PASS_STATUS = "Status"
    DETAIL_CSV_MESSAGE = "Message"
    DETAIL_CSV_FILE_NAME = "accuracy_checking_details"

    # result_csv
    RESULT_CSV_FORWARD_TEST_SUCCESS = "Forward Test Success"
    RESULT_CSV_BACKWARD_TEST_SUCCESS = "Backward Test Success"
    RESULT_CSV_FILE_NAME = "accuracy_checking_result"

    EPSILON = 1e-8

    class ProcessStatus:
        SUCCESS = "success"
        API_NOT_FOUND = "api_not_found"
        EXCEPTION_SKIP = "exception_skip"



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
KullbackLeiblerDivergence algorithm. This file mainly involves the compare function.
"""

import numpy as np

from msprobe.msaccucmp.algorithm_manager.algorithm_parameter import AlgorithmParameter
from msprobe.msaccucmp.cmp_utils import utils, log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager


FLOAT_EPSILON = np.finfo(np.float32).eps


def _normalized(dump_data: any) -> any:
    max_value = dump_data.max()
    min_value = dump_data.min()
    range_value = max_value - min_value
    if range_value != 0:
        dump_data_to_1 = (dump_data - min_value) / range_value
    else:
        dump_data_to_1 = dump_data
    # normalized, the sum of dump data is not equal with zero
    return np.maximum(dump_data_to_1, FLOAT_EPSILON)


def compare(my_output_dump_data: any, ground_truth_dump_data: any, args: AlgorithmParameter) -> (str, str):
    """
    compare my output dump data and the ground truth dump data
    by kullback leibler divergence
    1. P(x)=(x-Min)/(Max-Min)
    2. pdf = x/sum(x)
    3. sum(P(x)log(P(x)/Q(x)))
    :param my_output_dump_data: my output dump data to compare
    :param ground_truth_dump_data: the ground truth dump data to compare
    :param args: the algorithm parameter
    :return: the result of relative euclidean distance value and error message (the default is "")
    """
    my_output_dump_data_pdf = _normalized(my_output_dump_data)
    ground_true_dump_data_pdf = _normalized(ground_truth_dump_data)
    is_my_output_all_zero = np.all(my_output_dump_data_pdf == 0)
    is_ground_truth_all_zero = np.all(ground_true_dump_data_pdf == 0)
    if is_my_output_all_zero and is_ground_truth_all_zero:
        message = 'Cannot compare by KL Divergence. All the data is zero in %r and %r.' \
                  % (args.my_output_dump_file, args.ground_truth_dump_file)
        log.print_warn_log(message)
        return ConstManager.NAN, message
    if is_my_output_all_zero:
        message = 'Cannot compare by KL Divergence. All the data is zero in ' + args.my_output_dump_file + '.'
        log.print_warn_log(message)
        return ConstManager.NAN, message
    if is_ground_truth_all_zero:
        message = 'Cannot compare by KL Divergence. All the data is zero in ' + args.ground_truth_dump_file + '.'
        log.print_warn_log(message)
        return ConstManager.NAN, message

    norm_xx = my_output_dump_data_pdf / my_output_dump_data_pdf.sum()  # cannot be all 0
    norm_yy = ground_true_dump_data_pdf / ground_true_dump_data_pdf.sum()  # cannot be all 0
    result = (norm_xx * np.log(norm_xx / norm_yy)).sum()

    inf_message = ''
    if abs(result) < ConstManager.FLOAT_EPSILON:
        result = 0.0
    if str(result) == "inf":
        inf_message = 'Cannot compare by KL Divergence. The data contains 0 in %r.' % args.ground_truth_dump_file
    return utils.format_value(result), inf_message



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import namedtuple
import torch
import torch.nn as nn
import numpy as np

from einops import rearrange


from msprobe.pytorch.common.utils import logger

GTYPE = torch.float64  # arm host必须选择float64，x86环境选择float32即可，64也行。arm计算很慢，s=8k的场景建议使用x86
SOFTMAX_BUILD_MODE = "QKV"  # "MAX_SUM"

FaForwardParams = namedtuple("FaForwardParams",
                             ["q", "k", "v", "drop_mask", "attn_mask", "pse", "scalar_value", "keep_prob"])
FaBackwardParams = namedtuple("FaBackwardParams",
                              ["dx", "q", "k", "v", "softmax_res", "drop_mask", "pse", "scalar_value", "keep_prob"])
RebuildSoftmaxParams = namedtuple("RebuildSoftmaxParams",
                                  ["q", "k", "attn_mask", "pse", "scalar_value", "softmax_max", "softmax_sum"])


def softmax_forward(x):
    x_max = torch.max(x, dim=-1, keepdims=True)[0]
    x_sub = x.sub(x_max)
    y = torch.exp(x_sub)
    x_sum = y.sum(dim=-1, keepdims=True)
    res = y.div(x_sum)
    return res, x_max, x_sum


def softmax_grad(dp, softmax_res):
    muls = dp * softmax_res
    muls_r = muls.sum(dim=-1, keepdims=True)
    sub_r = dp - muls_r
    res = sub_r * softmax_res
    return res


def broadcast_kv(num_heads, num_kv_heads, kv_tensor, dtype):
    # 检查维度
    if kv_tensor.dim() != 4:
        raise ValueError(f"broadcast_kv: kv_tensor 必须是 4 维 (B, N_kv, S, D)，但得到 {kv_tensor.shape}")
    if num_kv_heads == 0 or num_kv_heads > num_heads:
        raise ValueError("broadcast_kv: num_kv_heads 必须大于 0 且不超过 num_heads。")
    if num_heads % num_kv_heads != 0:
        raise ValueError(f"broadcast_kv: num_heads({num_heads}) 必须能被 num_kv_heads({num_kv_heads}) 整除。")


    factor = num_heads // num_kv_heads
    kv_shape = kv_tensor.shape
    b = kv_shape[0]
    s = kv_shape[2]
    d = kv_shape[3]
    kv_res = torch.zeros([b, num_heads, s, d]).to(dtype)
    for i in range(num_heads):
        j = i // factor
        kv_res[:, i:i + 1, :, :] = kv_tensor[:, j:j + 1, :, :]
    return kv_res


def calculate_qk(q, k, attn_mask, pse, scalar_value):
    # 基本形状检查
    if q.dim() < 4 or k.dim() < 4:
        raise ValueError(f"calculate_qk: q,k 必须至少 4 维，q={q.dim()}，k={k.dim()}")
    # 检查 head_dim 一致性
    if q.size(-1) != k.size(-1):
        raise ValueError(f"calculate_qk: q.head_dim({q.size(-1)}) != k.head_dim({k.size(-1)})")

    if k.dim() != 4:
        raise ValueError(f"k tensor dimension must be 4, but got {k.dim()} dimensions (shape: {k.shape})")

    if k.dim() == 3:
        k = k.unsqueeze(1)  # 在head维度扩展

    if pse is None or len(pse.shape) == 0:
        qk = torch.matmul(q, k.permute(0, 1, 3, 2)).mul(scalar_value)
    else:
        qk = (torch.matmul(q, k.permute(0, 1, 3, 2)) + pse).mul(scalar_value)
    if attn_mask is None or len(attn_mask.shape) == 0:
        return qk
    else:
        qk = qk + attn_mask.bool() * (-40000.0)  # -10000
    return qk


def fusion_attention_forward(forward_params):
    q = forward_params.q
    k = forward_params.k
    v = forward_params.v
    drop_mask = forward_params.drop_mask
    attn_mask = forward_params.attn_mask
    pse = forward_params.pse
    scalar_value = forward_params.scalar_value
    keep_prob = forward_params.keep_prob

    # 拦截 keep_prob 为 0 的情况，防止除零
    if keep_prob == 0:
        raise ValueError("fusion_attention_forward: keep_prob 不能为 0，避免除零错误。")

    qk = calculate_qk(q, k, attn_mask, pse, scalar_value)
    softmax_res, softmax_max, softmax_sum = softmax_forward(qk)
    if drop_mask is None or len(drop_mask.shape) == 0:
        drop_res = softmax_res
    else:
        drop_res = softmax_res * drop_mask * (1.0 / keep_prob)
    y = torch.matmul(drop_res, v)
    return y, softmax_max, softmax_sum


def fusion_attention_backward(backward_params):
    dx = backward_params.dx
    q = backward_params.q
    k = backward_params.k
    v = backward_params.v
    softmax_res = backward_params.softmax_res
    drop_mask = backward_params.drop_mask
    pse = backward_params.pse
    scalar_value = backward_params.scalar_value
    keep_prob = backward_params.keep_prob

    # 拦截 keep_prob 为 0 的情况，防止除零
    if keep_prob == 0:
        raise ValueError("fusion_attention_backward: keep_prob 不能为 0，避免除零错误。")

    dp = torch.matmul(dx, v.permute(0, 1, 3, 2))
    if drop_mask is None or len(drop_mask.shape) == 0:
        drop_res = softmax_res.permute(0, 1, 3, 2)
        dp_drop = dp
    else:
        drop_res = softmax_res.mul(drop_mask).mul(1.0 / keep_prob).permute(0, 1, 3, 2)
        dp_drop = dp * drop_mask * (1.0 / keep_prob)
    dv = torch.matmul(drop_res, dx)
    softmax_grad_res = (softmax_grad(dp_drop, softmax_res) * scalar_value)
    dq = torch.matmul(softmax_grad_res, k)
    dk = torch.matmul(softmax_grad_res.permute(0, 1, 3, 2), q)
    return dq, dk, dv


def parse_bsnd_args(query, key, head_num, input_layout):
    supported_input_layout = ["BSH", "SBH", "BSND", "BNSD", "TND"]
    b, s1, s2, n1, n2, d, h1, h2 = None, None, None, head_num, None, None, None, None

    if not isinstance(input_layout, str) or input_layout not in supported_input_layout:
        raise ValueError(f"Invalid input_layout arg which must be one of {supported_input_layout}.")

    if input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")

    # 防止 head_num 为 0
    if n1 == 0:
        raise ValueError("parse_bsnd_args: head_num (n1) 不能为 0，避免除零错误。")

    try:
        if input_layout == "BSH":
            b, s1, h1 = query.shape
            _, s2, h2 = key.shape
            d = h1 // n1
            # 拦截 d 为 0 的情况
            if d == 0:
                raise ValueError("parse_bsnd_args: 计算得到的 head_dim d 不能为 0。")
            n2 = h2 // d
        elif input_layout == "SBH":
            s1, b, h1 = query.shape
            s2, _, h2 = key.shape
            d = h1 // n1
            if d == 0:
                raise ValueError("parse_bsnd_args: 计算得到的 head_dim d 不能为 0。")
            n2 = h2 // d
        elif input_layout == "BSND":
            b, s1, n1, d = query.shape
            _, s2, n2, _ = key.shape
            if d == 0:
                raise ValueError("parse_bsnd_args: head_dim d 不能为 0。")
            h1 = n1 * d
            h2 = n2 * d
        elif input_layout == "BNSD":
            b, n1, s1, d = query.shape
            _, n2, s2, _ = key.shape
            if d == 0:
                raise ValueError("parse_bsnd_args: head_dim d 不能为 0。")
            h1 = n1 * d
            h2 = n2 * d
    except Exception as e:
        raise ValueError(f"query.shape: {query.shape}, key.shape: {key.shape}, parse_bsnd_args error: {e}") from e

    ret = (b, s1, s2, n1, n2, d, h1, h2, query.dtype)
    return ret


def convert_from_bnsd(_input, input_layout):
    """
    transform qkv from bnsd to input_layout.
    B: batch_size
    S: sequence_length
    N: num_heads
    D: head_dim
    Args:
       _input (torch.Tensor): tensor of shape (B,N,S,D)
        input_layout (str): "BSH" or "SBH" or "BSND" or "BNSD" or "TND"
    Returns:
        tensor of shape (B,N,S,D) or (B,S,N,D) or (S,B,H) or (B,S,H)
    """
    if input_layout == "BSH":
        # (B,N,S,D)=>(B,S,N*D)
        out = rearrange(_input, 'b n s d -> b s (n d)').contiguous()
    elif input_layout == "SBH":
        # (B,N,S,D)=>(S,B,N*D)
        out = rearrange(_input, 'b n s d -> s b (n d)').contiguous()
    elif input_layout == "BSND":
        # (B,N,S,D)=>(B,S,N,D)
        out = rearrange(_input, 'b n s d -> b s n d').contiguous()
    elif input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    else:
        out = _input
    return out


def convert_to_bnsd(_input, n, input_layout):
    """
    transform qkv from input_layout to bnsd.
    B: batch_size
    S: sequence_length
    N: num_heads
    D: head_dim
    Args:
        _input (torch.Tensor): tensor of shape (B,N,S,D) or (B,S,N,D) or (S,B,H) or (B,S,H)
        n (int): num_heads
        input_layout (str):"BSH" or "SBH" or "BSND" or "BNSD" or "TND"
    Returns:
        tensor of shape (B,N,S,D)
    """
    if input_layout == "BSH":
        # (B,S,N*D)=>(B,N,S,D)
        out = rearrange(_input, 'b s (n d) -> b n s d', n=n)
    elif input_layout == "SBH":
        # (S,B,N*D)=>(B,N,S,D)
        out = rearrange(_input, 's b (n d) -> b n s d', n=n)
    elif input_layout == "BSND":
        # (B,S,N,D)=>(B,N,S,D)
        out = rearrange(_input, 'b s n d -> b n s d', n=n)
    elif input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    else:
        out = _input
    if out.dim() != 4:
        raise ValueError(f"convert qkv format failed with input_layout {input_layout}.")
    return out.to(GTYPE)


def generate_attn_mask(*args):
    """
    # 当sparse_mode=2、3、4时小算子到融合算子会走这个优化，反过来看就要拆解回原来的基本实现
    ===> attn_mask = torch.from_numpy(np.triu(np.ones([2048, 2048]), k=1)).to(dtype)
    """

    sparse_mode, attn_mask, b, n1, s1, s2, pre_tocken, next_tocken, dtype = args
    shape = [s1, s2]

    if attn_mask is not None:
        # 当FA的输入已经包含attn_mask时，可以认为已经是转换之后的mask矩阵了，有三种特殊场景，即稀疏矩阵场景，需要进行逆向还原
        if sparse_mode == 2 or sparse_mode == 3 or sparse_mode == 4:
            logger.info(f"s1: {s1}, s2:{s2}, attn_mask.shape:{attn_mask.shape}, attn_mask.dtype:{attn_mask.dtype}")

            if attn_mask.dim() == 2 and attn_mask.shape[0] == 2048 and attn_mask.shape[1] == 2048:
                if attn_mask.equal(torch.from_numpy(np.triu(np.ones([2048, 2048]), k=1)).to(attn_mask.dtype)):
                    if sparse_mode == 2:
                        attn_mask = torch.from_numpy(np.triu(np.ones(shape), k=1))
                    elif sparse_mode == 3:
                        attn_mask = torch.from_numpy(np.triu(np.ones(shape), k=s2 - s1 + 1))
                    elif sparse_mode == 4:
                        attn_mask_u = torch.from_numpy(np.triu(np.ones(shape), k=next_tocken + 1))
                        attn_mask_l = torch.from_numpy(np.tril(np.ones(shape), k=-pre_tocken - 1))
                        attn_mask = attn_mask_u + attn_mask_l
                    logger.debug(f"反向转换attn_mask {attn_mask.shape}")
                    return attn_mask.to(dtype)

        return attn_mask.to(dtype)

    if attn_mask is not None:
        if attn_mask.dim() == 2:
            if attn_mask.shape[0] != s1 or attn_mask.shape[1] != s2:
                raise ValueError(f"Invalid attn_mask shape `SS` {attn_mask.shape}")
            shape = [s1, s2]
        elif attn_mask.dim() == 4:
            if attn_mask.shape[1] == 1:
                shape = [b, 1, s1, s2] if b != 1 else [1, 1, s1, s2]
            else:
                shape = [b, n1, s1, s2] if b != 1 else [1, n1, s1, s2]

    if sparse_mode == 0:
        attn_mask_u = torch.from_numpy(np.triu(np.ones(shape), k=next_tocken + 1))
        attn_mask_l = torch.from_numpy(np.tril(np.ones(shape), k=-pre_tocken - 1))
        attn_mask = attn_mask_u + attn_mask_l
    elif sparse_mode == 1:  # no sparse
        attn_mask = torch.from_numpy(np.zeros(shape))
    elif sparse_mode == 2:
        attn_mask = torch.from_numpy(np.triu(np.ones(shape), k=1))
    elif sparse_mode == 3:
        attn_mask = torch.from_numpy(np.triu(np.ones(shape), k=s2 - s1 + 1))
    elif sparse_mode == 4:
        attn_mask_u = torch.from_numpy(np.triu(np.ones(shape), k=next_tocken + 1))
        attn_mask_l = torch.from_numpy(np.tril(np.ones(shape), k=-pre_tocken - 1))
        attn_mask = attn_mask_u + attn_mask_l
    # 注:不会出现sparse_mode=5的情况，该情况要求必须要传入attn_mask，且attn_mask矩阵数据格式须为BNSS或B1SS，
    # 因此可以认为FA的输入已经是正确的attn_mask了
    return attn_mask.to(dtype)


def generate_kv(key, value, n1, n2):
    # N不等长适配by cdy
    if not (n1 == n2):
        k_new = broadcast_kv(n1, n2, key, key.dtype)
        v_new = broadcast_kv(n1, n2, value, value.dtype)
    else:
        k_new = key
        v_new = value
    return k_new, v_new


def rebuid_softmax_by_qkv(q, k, attn_mask, pse, scalar_value):
    """
    attention = softmax(QK^T/sqrt(d))V
    softmax(x_i) = e^(x_i - x_max) / sum(e^(x_i - x_max))
    """
    logger.info("Using QKV to rebuild original softmax")
    qk = calculate_qk(q, k, attn_mask, pse, scalar_value)
    softmax_res, _, _ = softmax_forward(qk)
    return softmax_res


def rebuild_softmax_by_max_sum(softmax_params):
    """
    attention = softmax(QK^T/sqrt(d))V
    softmax(x_i) = e^(x_i - x_max_i) / x_sum_i)
    """
    q = softmax_params.q
    k = softmax_params.k
    attn_mask = softmax_params.attn_mask
    pse = softmax_params.pse
    scalar_value = softmax_params.scalar_value
    softmax_max = softmax_params.softmax_max
    softmax_sum = softmax_params.softmax_sum
    logger.info("Using softmax_max and softmax_sum to rebuild original softmax")

    qk = calculate_qk(q, k, attn_mask, pse, scalar_value)
    if softmax_max.shape[-1] == 0:
        raise ValueError(f"softmax_max.shape[-1] must be non-zero, softmax_max.shape: {softmax_max.shape}")
    repeat_dim = qk.shape[-1] // softmax_max.shape[-1]
    softmax_res = torch.exp(qk.sub(softmax_max.repeat(1, 1, 1, repeat_dim))).div(
        softmax_sum.repeat(1, 1, 1, repeat_dim))
    return softmax_res


def get_head_num(*args, **kwargs):
    if kwargs.get("head_num", None):
        head_num = kwargs.get("head_num")
    elif len(args) >= 4:
        head_num = args[3]
    else:
        raise ValueError(f"Unsupported npu_fusion_attention args {args}.")
    return head_num


def get_input_layout(*args, **kwargs):
    if kwargs.get("input_layout", None):
        input_layout = kwargs.get("input_layout")
    elif len(args) >= 5:
        input_layout = args[4]
    else:
        raise ValueError(f"Unsupported npu_fusion_attention args {args}.")
    return input_layout


def npu_fusion_attention_forward_patch(*args, **kwargs):
    if len(args) < 2:
        raise RuntimeError("npu_fusion_attention_forward_patch: length of args should be greater than or equal to 2.")

    # query, key, value, head_num, input_layout
    head_num = get_head_num(*args, **kwargs)
    input_layout = get_input_layout(*args, **kwargs)

    b, s1, s2, n1, n2, d, h1, h2, dtype = parse_bsnd_args(args[0], args[1], head_num, input_layout)
    # 此处 d 已在 parse_bsnd_args 中检查为非零
    if n1 == n2 and s1 == s2:
        logger.debug(f"running case : BNSD = {b}_{n1}_{s1}_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    else:
        logger.debug(f"running case: BNSD = {b}_{n1}({n2})_{s1}({s2})_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    if n2 == 0:
        raise ValueError("n2 不能为 0，避免除零错误。")
    if not (n1 % n2 == 0 and n1 >= n2):
        raise ValueError(f"N1与N2不匹配,请检查: n1 = {n1}, n2 = {n2}.")

    dims_kwargs = {
        "b": b, "s1": s1, "s2": s2, "n1": n1, "n2": n2,
        "d": d, "h1": h1, "h2": h2, "dtype": dtype
    }
    new_kwargs = {
        "keep_prob": 1,  # 注意：如果外部传入 keep_prob 为 0，也会在 fusion_attention_forward 中捕获
        "scalar_value": kwargs.get("scalar_value", 1 / (d ** 0.5)),
        "sparse_mode": kwargs.get("sparse_mode", 0),
        "prefix": kwargs.get("prefix"),
        "pre_tockens": kwargs.get("pre_tockens", 2147483647),
        "next_tockens": kwargs.get("next_tockens", 2147483647),
        "pse": kwargs.get("pse"),
        "padding_mask": kwargs.get("padding_mask"),
        "attn_mask": kwargs.get("attn_mask")
    }

    return args, dims_kwargs, new_kwargs


def npu_fusion_attention_backward_patch(*args, **kwargs):
    if len(args) != 6:
        raise ValueError(f"Unsupported npu_fusion_attention_grad args {args}.")

    b, s1, s2, n1, n2, d, h1, h2, dtype = parse_bsnd_args(args[0], args[1], args[4], args[5])
    # 此处 d 已在 parse_bsnd_args 中检查为非零
    if n1 == n2 and s1 == s2:
        logger.info(f"running case : bnsd = {b}_{n1}_{s1}_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    else:
        logger.info(f"running case: bnsd = {b}_{n1}({n2})_{s1}({s2})_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    if n2 == 0:
        raise ValueError("n2 不能为 0，避免除零错误。")
    if not (n1 % n2 == 0 and n1 >= n2):
        raise ValueError(f"N1与N2不匹配,请检查: n1 = {n1}, n2 = {n2}.")

    dims_kwargs = {
        "b": b, "s1": s1, "s2": s2, "n1": n1, "n2": n2,
        "d": d, "h1": h1, "h2": h2, "dtype": dtype
    }

    new_kwargs = {
        "keep_prob": 1,  # 同上，fusion_attention_backward 内会拦截 keep_prob 为 0 的情况
        "scalar_value_value": kwargs.get("scalar_value_value", 1 / (d ** 0.5)),
        "sparse_mode": kwargs.get("sparse_mode", 0),
        "prefix": kwargs.get("prefix"),
        "pre_tockens": kwargs.get("pre_tockens", 2147483647),
        "next_tockens": kwargs.get("next_tockens", 2147483647),
        "pse": kwargs.get("pse"),
        "padding_mask": kwargs.get("padding_mask"),
        "softmax_max": kwargs.get("softmax_max"),
        "softmax_sum": kwargs.get("softmax_sum"),
        "softmax_in": kwargs.get("softmax_in"),
        "attention_in": kwargs.get("attention_in"),
        "seed": kwargs.get("seed", 0),
        "offset": kwargs.get("offset", 0),
        "numels": kwargs.get("numels", 0),
        "attn_mask": kwargs.get("attn_mask")
    }

    return args, dims_kwargs, new_kwargs


class FlashAttentionScore(nn.Module):
    def __init__(self):
        super(FlashAttentionScore, self).__init__()
        # You can initialize any parameters here if necessary

    def forward(self, *inputs, **kwargs):
        # Extract the inputs for the attention calculation
        new_args, dims_kwargs, new_kwargs = npu_fusion_attention_forward_patch(*inputs, **kwargs)
        query, key, value = new_args[0], new_args[1], new_args[2]

        input_layout = get_input_layout(*inputs, **kwargs)

        n1 = dims_kwargs.get("n1")
        n2 = dims_kwargs.get("n2")
        s1 = dims_kwargs.get("s1")
        s2 = dims_kwargs.get("s2")
        b = dims_kwargs.get("b")
        dtype = dims_kwargs.get("dtype")
        attn_mask = new_kwargs.get("attn_mask")
        keep_prob = new_kwargs.get("keep_prob")
        sparse_mode = new_kwargs.get("sparse_mode")
        pre_tockens = new_kwargs.get("pre_tockens")
        next_tockens = new_kwargs.get("next_tokens")
        pse = new_kwargs.get("real_shift")
        scalar_value = new_kwargs.get("scalar_value")

        args_temp = [sparse_mode, attn_mask, b, n1, s1, s2, pre_tockens, next_tockens, dtype]

        attn_mask = generate_attn_mask(*args_temp)
        query = convert_to_bnsd(query, n1, input_layout)
        key = convert_to_bnsd(key, n2, input_layout)
        value = convert_to_bnsd(value, n2, input_layout)

        forward_params = FaForwardParams(
            q=query,
            k=key,
            v=value,
            drop_mask=None,
            attn_mask=attn_mask,
            pse=pse,
            scalar_value=scalar_value,
            keep_prob=keep_prob
        )

        out_golden, softmax_max, softmax_sum = fusion_attention_forward(forward_params)

        # If output dimension is 5, reshape accordingly
        if out_golden.dim() == 5:
            out_golden = out_golden.reshape(out_golden.size(0),
                                            out_golden.size(1) * out_golden.size(2),
                                            out_golden.size(3), out_golden.size(4))

        out_golden = convert_from_bnsd(out_golden, input_layout)

        # Ensure the output matches the desired layout
        out_golden = out_golden.cpu(), softmax_max.repeat(1, 1, 1, 8).cpu(), softmax_sum.repeat(1, 1, 1, 8).cpu()

        return out_golden

    def backward(self, *inputs, **kwargs):
        # The backward pass will be similar to what was described for the gradient computation
        new_args, dims_kwargs, new_kwargs = npu_fusion_attention_backward_patch(*inputs, **kwargs)
        query, key, value, dx, input_layout = new_args[0], new_args[1], new_args[2], new_args[3], new_args[5]
        n1 = dims_kwargs.get("n1")
        n2 = dims_kwargs.get("n2")
        s1 = dims_kwargs.get("s1")
        s2 = dims_kwargs.get("s2")
        b = dims_kwargs.get("b")
        dtype = dims_kwargs.get("dtype")
        attn_mask = new_kwargs.get("attn_mask")
        keep_prob = new_kwargs.get("keep_prob")
        sparse_mode = new_kwargs.get("sparse_mode")
        pre_tockens = new_kwargs.get("pre_tockens")
        next_tockens = new_kwargs.get("next_tockens")
        pse = new_kwargs.get("pse")
        softmax_max = new_kwargs.get("softmax_max")
        softmax_sum = new_kwargs.get("softmax_sum")
        scalar_value = new_kwargs.get("scalar_value")

        args_temp = [sparse_mode, attn_mask, b, n1, s1, s2, pre_tockens, next_tockens, dtype]
        attn_mask = generate_attn_mask(*args_temp)

        query = convert_to_bnsd(query, n1, input_layout)
        dx = convert_to_bnsd(dx, n1, input_layout)
        key = convert_to_bnsd(key, n2, input_layout)
        value = convert_to_bnsd(value, n2, input_layout)

        k_new, v_new = generate_kv(key, value, n1, n2)

        if SOFTMAX_BUILD_MODE == "QKV":
            softmax_res = rebuid_softmax_by_qkv(query, k_new, attn_mask, pse, scalar_value)
        else:
            softmax_params = RebuildSoftmaxParams(query, k_new, attn_mask, pse, scalar_value, softmax_max, softmax_sum)
            softmax_res = rebuild_softmax_by_max_sum(softmax_params)

        backward_params = FaBackwardParams(dx, query, k_new, v_new, softmax_res, None, pse, scalar_value, keep_prob)
        dq, dk, dv = fusion_attention_backward(backward_params)

        # Reshape as needed
        if dq.dim() == 5:
            dq = dq.reshape(dq.size(0), dq.size(1) * dq.size(2), dq.size(3), dq.size(4))
        if dk.dim() == 5:
            dk = dk.reshape(dk.size(0), dk.size(1) * dk.size(2), dk.size(3), dk.size(4))
        if dv.dim() == 5:
            dv = dv.reshape(dv.size(0), dv.size(1) * dv.size(2), dv.size(3), dv.size(4))

        dq = convert_from_bnsd(dq, input_layout)
        dk = convert_from_bnsd(dk, input_layout)
        dv = convert_from_bnsd(dv, input_layout)

        return dq.cpu(), dk.cpu(), dv.cpu()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
convert format from NHWC to FRACTAL_Z.
"""
from itertools import product
import numpy as np

from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.utils import least_common_multiple as lcm
from msprobe.msaccucmp.cmp_utils.utils import ceiling_divide as ceil


def _get_axis(gnh_axis: list, value_map: dict, dst_c: int, w_axis: int) -> int:
    g_axis = gnh_axis[0]
    h_axis = gnh_axis[2]
    kh_axis = value_map.get('kh_axis')
    kw_axis = value_map.get('kw_axis')
    e_multi = value_map.get('e_multi')
    tmp_value = (g_axis // e_multi) * (dst_c // ConstManager.C0_AXIS) * kh_axis * kw_axis
    return tmp_value + (dst_c // ConstManager.C0_AXIS) * kh_axis * kw_axis + h_axis * kw_axis + w_axis


def _get_count_for_axis(shape_from: list, g_num: int, e_multi: int) -> int:
    c_ori = shape_from[3]
    kh_axis = shape_from[1]
    kw_axis = shape_from[2]
    c_opt = ceil(e_multi * c_ori, ConstManager.C0_AXIS) * ConstManager.C0_AXIS
    c1_axis = ceil(c_opt, ConstManager.C0_AXIS)
    return g_num * c1_axis * kh_axis * kw_axis


def convert(shape_from: list, shape_to: list, array: any, group: int = 1) -> any:
    """
    Convert the data format from NHWC to FRACTAL_Z
    :param shape_from: the shape before convert
    :param shape_to: the shape after convert
    :param array: the one-dimensional array
    :param group: group for group_conv, default value is 1
    :return: the data array of FRACTAL_Z shape
    """
    _ = shape_to
    # get before convert shape
    kh_axis = shape_from[1]
    kw_axis = shape_from[2]
    c_ori = shape_from[3]
    array_shape = array.reshape(shape_from[0], kh_axis, kw_axis, c_ori)

    # CUBE Unit K and N
    n_ori = shape_from[0] // group

    # Specific multiplying algorithm to get e_multi
    e_multi = min(lcm(lcm(c_ori, ConstManager.C0_AXIS) // c_ori, lcm(n_ori, ConstManager.N0_AXIS) // n_ori), group)
    array_to = np.zeros(
        (_get_count_for_axis(shape_from, ceil(group, e_multi), e_multi),
         ceil(ceil(e_multi * n_ori, ConstManager.N0_AXIS) * ConstManager.N0_AXIS, ConstManager.N0_AXIS),
         ConstManager.N0_AXIS, ConstManager.C0_AXIS), dtype=array_shape.dtype)
    # convert nhwc to gc1hwn1n0c0
    for g_axis, n_axis, h_axis, w_axis, c_axis in \
            product(range(group), range(n_ori), range(kh_axis), range(kw_axis), range(c_ori)):
        e_val = g_axis % e_multi
        dst_c = e_val * c_ori + c_axis
        dst_n = e_val * n_ori + n_axis
        src_n = g_axis * n_ori + n_axis
        array_to[_get_axis([g_axis, n_axis, h_axis],
                            {'e_multi': e_multi, 'kh_axis': kh_axis, 'kw_axis': kw_axis},
                            dst_c, w_axis)][dst_n // ConstManager.N0_AXIS][
            dst_n % ConstManager.N0_AXIS][dst_c % ConstManager.C0_AXIS] = \
            array_shape[src_n][h_axis][w_axis][c_axis]
    return array_to




# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.compare.merge_result.merge_result import merge_result


def _merge_result_parser(parser):
    parser.add_argument("-i", "--input_dir", dest="input_dir", type=str,
                        help="<Required> The compare result path, a dir.", required=True)
    parser.add_argument("-o", "--output_dir", dest="output_dir", type=str,
                        help="<Required> The result merge output path, a dir.", required=True)
    parser.add_argument("-config", "--config-path", dest="config_path", type=str,
                        help="<Required> Yaml path containing distribute APIs and compare indexes for merging data "
                             "from compare results.",
                        required=True)


def merge_result_cli(args):
    merge_result(args.input_dir, args.output_dir, args.config_path)



#!/usr/bin/env python3
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Utility helpers that support the torchair accuracy comparison flow.
"""

import datetime
import os
from typing import Any, Dict, List

import numpy as np
import pandas as pd

from msprobe.core.common.const import FileCheckConst
from msprobe.core.common.file_utils import FileChecker, create_directory, write_df_to_csv
from msprobe.core.common.log import logger
from msprobe.infer.utils.check.rule import Rule
from msprobe.infer.utils.constants import TENSOR_MAX_SIZE
from msprobe.infer.utils.util import safe_torch_load

GLOBAL_HISTORY_AIT_DUMP_PATH_LIST = ["msit_dump", "ait_dump"]

TOKEN_ID = "token_id"
DATA_ID = "data_id"
GOLDEN_DATA_PATH = "golden_data_path"
GOLDEN_DTYPE = "golden_dtype"
GOLDEN_SHAPE = "golden_shape"
GOLDEN_MAX_VALUE = "golden_max_value"
GOLDEN_MIN_VALUE = "golden_min_value"
GOLDEN_MEAN_VALUE = "golden_mean_value"
GOLDEN_OP_TYPE = "golden_op_type"
MY_DATA_PATH = "my_data_path"
MY_DTYPE = "my_dtype"
MY_SHAPE = "my_shape"
MY_MAX_VALUE = "my_max_value"
MY_MIN_VALUE = "my_min_value"
MY_MEAN_VALUE = "my_mean_value"
MY_OP_TYPE = "my_op_type"
CMP_FAIL_REASON = "cmp_fail_reason"
DTYPE = "dtype"
SHAPE = "shape"
MAX_VALUE = "max_value"
MIN_VALUE = "min_value"
MEAN_VALUE = "mean_value"

CSV_GOLDEN_HEADER_BASE = [
    TOKEN_ID,
    DATA_ID,
    GOLDEN_DATA_PATH,
    GOLDEN_OP_TYPE,
    GOLDEN_DTYPE,
    GOLDEN_SHAPE,
    GOLDEN_MAX_VALUE,
    GOLDEN_MIN_VALUE,
    GOLDEN_MEAN_VALUE,
    MY_DATA_PATH,
    MY_OP_TYPE,
    MY_DTYPE,
    MY_SHAPE,
    MY_MAX_VALUE,
    MY_MIN_VALUE,
    MY_MEAN_VALUE,
]


def _lazy_import_torch():
    try:
        import torch  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as err:
        if err.name == "torch":
            logger.error("Missing dependency: torch. Please install torch to use torchair compare.")
        raise
    return torch


def _lazy_cmp_alg_maps():
    from msprobe.infer.utils import cmp_algorithm  # pylint: disable=import-outside-toplevel

    return cmp_algorithm.CMP_ALG_MAP, cmp_algorithm.CUSTOM_ALG_MAP


def _build_csv_golden_header():
    cmp_alg_map, _ = _lazy_cmp_alg_maps()
    header = list(CSV_GOLDEN_HEADER_BASE)
    header.extend(list(cmp_alg_map.keys()))
    header.append(CMP_FAIL_REASON)
    return header


class BasicDataInfo:
    """Metadata holder for a pair of tensors."""

    count_data_id = 0  # Count data_id, increment by 1 every time creating a new instance
    TORCH_UNSUPPORTED_D_TYPE_MAP = {"uint16": "int32", "uint32": "int64"}

    def __init__(self, golden_data_path, my_data_path, token_id=None, data_id=None, op_type=None):
        global_path = os.path.realpath(golden_data_path.split(",")[0])
        self._check_path(global_path)
        my_real_path = os.path.realpath(my_data_path.split(",")[0])
        self._check_path(my_real_path)

        self.my_data_path, self.golden_data_path = os.path.realpath(my_data_path), os.path.realpath(golden_data_path)
        self.token_id = self.get_token_id(self.my_data_path) if token_id is None else token_id
        self.data_id = self.count_data_id if data_id is None else data_id
        self.my_op_type, self.golden_op_type = self._validate_op_type(op_type)
        self._count()

    @staticmethod
    def _check_path(path):
        path_type = FileCheckConst.DIR if os.path.isdir(path) else FileCheckConst.FILE
        FileChecker(path, path_type, ability=FileCheckConst.READ_ABLE).common_check()

    @staticmethod
    def _validate_op_type(op_type):
        if op_type is None:
            return None, None

        if isinstance(op_type, (list, tuple)) and len(op_type) == 2:
            return op_type[0], op_type[1]
        raise ValueError("op_type must be a list or tuple containing two elements")

    @classmethod
    def _count(cls):
        cls.count_data_id += 1

    def to_dict(self):
        return {
            TOKEN_ID: str(self.token_id),
            DATA_ID: str(self.data_id),
            GOLDEN_DATA_PATH: self.golden_data_path,
            GOLDEN_OP_TYPE: self.golden_op_type,
            MY_DATA_PATH: self.my_data_path,
            MY_OP_TYPE: self.my_op_type,
        }

    def get_token_id(self, cur_path):
        dump_filename_idx = 4
        dump_tensor_idx = 3
        dirseg = cur_path.split(os.path.sep)
        if len(dirseg) > 16:
            raise RecursionError(f'The depth of "{cur_path}" directory is too deep.')
        if len(dirseg) < dump_filename_idx:
            return 0
        flag1 = dirseg[-dump_tensor_idx] in {"tensors", "torch_tensors"}
        flag2 = any(dirseg[-dump_filename_idx].startswith(x) for x in GLOBAL_HISTORY_AIT_DUMP_PATH_LIST)
        if flag1 and flag2:
            try:
                token_id = int(dirseg[-1])
            except (IndexError, AttributeError, TypeError, ValueError) as err:
                logger.debug(f"get_token_id error, dirseg: {dirseg}, error: {err}")
                token_id = 0
        else:
            token_id = self.get_token_id(os.path.dirname(cur_path))
        return token_id


def fill_row_data(
    data_info: BasicDataInfo,
    loaded_my_data: Any = None,
    loaded_golden_data: Any = None,
):
    golden_data_path, my_data_path = data_info.golden_data_path, data_info.my_data_path
    row_data = data_info.to_dict()
    if loaded_golden_data is None and not os.path.isfile(golden_data_path):
        row_data[CMP_FAIL_REASON] = f"golden_data_path: {golden_data_path} is not a file."
        return row_data
    if loaded_my_data is None and not os.path.isfile(my_data_path):
        row_data[CMP_FAIL_REASON] = f"my_data_path: {my_data_path} is not a file."
        return row_data
    golden_data = load_as_torch_tensor(golden_data_path, loaded_golden_data)
    my_data = load_as_torch_tensor(my_data_path, loaded_my_data)

    compare_metrics_dict = compare_data(golden_data, my_data)
    tensor_basic_info_dict = set_tensor_basic_info_in_row_data(golden_data, my_data)
    row_data.update(compare_metrics_dict)
    row_data.update(tensor_basic_info_dict)

    return row_data


def load_as_torch_tensor(data_path, loaded_data=None):
    torch = _lazy_import_torch()
    if loaded_data is not None:
        if str(loaded_data.dtype) in BasicDataInfo.TORCH_UNSUPPORTED_D_TYPE_MAP:
            mapped = BasicDataInfo.TORCH_UNSUPPORTED_D_TYPE_MAP.get(str(loaded_data.dtype))
            loaded_data = loaded_data.astype(mapped)
        return loaded_data if isinstance(loaded_data, torch.Tensor) else torch.from_numpy(loaded_data)
    return read_data(data_path)


def get_tensor_basic_info(data: Any) -> Dict[str, Any]:
    tensor_info: Dict[str, Any] = {}
    tensor_info[DTYPE] = str(data.dtype)
    tensor_info[SHAPE] = str(list(data.shape))
    if 0 not in data.shape:
        data = data.float()
        tensor_info[MAX_VALUE] = data.max().item()
        tensor_info[MIN_VALUE] = data.min().item()
        tensor_info[MEAN_VALUE] = data.mean().item()
    return tensor_info


def set_tensor_basic_info_in_row_data(golden_data, my_data):
    row_data: Dict[str, Any] = {}
    golden_info = get_tensor_basic_info(golden_data)
    row_data.update({f"golden_{key}": value for key, value in golden_info.items()})

    my_info = get_tensor_basic_info(my_data)
    row_data.update({f"my_{key}": value for key, value in my_info.items()})
    return row_data


def save_compare_result_to_csv(gathered_row_data, output_path=".", columns=None, rank_id=-1):
    create_directory(output_path)
    cur_time = datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%d%H%M%S")
    columns = columns or _build_csv_golden_header()
    if rank_id != -1:
        csv_save_path = os.path.join(output_path, f"msit_cmp_report_rank{rank_id}_{cur_time}.csv")
    else:
        csv_save_path = os.path.join(output_path, f"msit_cmp_report_{cur_time}.csv")
    gathered_row_data = list(
        filter(
            lambda item: not (CMP_FAIL_REASON in item and item[CMP_FAIL_REASON] == "data shape doesn't match."),
            gathered_row_data,
        )
    )
    for row_data in gathered_row_data[:]:
        if GOLDEN_DTYPE in row_data and MY_DTYPE in row_data:
            if (row_data[GOLDEN_DTYPE] == "torch.int8") ^ (row_data[MY_DTYPE] == "torch.int8"):
                gathered_row_data.remove(row_data)

    data_frame = pd.DataFrame(gathered_row_data, columns=columns)
    data_frame.fillna(value="", inplace=True)
    data_frame.dropna(axis=0, how="all", inplace=True)
    write_df_to_csv(data_frame, csv_save_path)
    logger.info(f"Saved comparing results: {csv_save_path}")
    return csv_save_path


def compare_data(golden_data, my_data):
    torch = _lazy_import_torch()
    if not hasattr(compare_data, "index"):
        compare_data.index = 0

    golden_data_dtype = golden_data.dtype
    my_data_dtype = my_data.dtype
    if golden_data_dtype != torch.float32:
        logger.debug(
            f"The dtype of golden_data with index {compare_data.index} is {golden_data_dtype}, convert it to fp32"
        )
    if my_data_dtype != torch.float32:
        logger.debug(f"The dtype of my_data with index {compare_data.index} is {my_data_dtype}, convert it to fp32")
    golden_data_fp32 = golden_data.reshape(-1).float()
    my_data_fp32 = my_data.reshape(-1).float()
    compare_data.index += 1
    compare_metrics_dict = compare_tensor(golden_data_fp32, my_data_fp32)
    return compare_metrics_dict


def read_data(data_path):
    torch = _lazy_import_torch()
    data_path = os.path.realpath(data_path)
    Rule.input_file().check(data_path, will_raise=True)
    if data_path.endswith(".npy"):
        data = torch.as_tensor(np.load(data_path, allow_pickle=False))
    elif data_path.endswith((".pth", ".pt")):
        data = safe_torch_load(data_path, map_location=torch.device("cpu"))
    else:
        logger.error(f"Unsupported data format {data_path}")
        raise TypeError("Unsupported data format.")

    if isinstance(data, torch.Tensor):
        return data.cpu()
    if isinstance(data, np.ndarray):
        return torch.from_numpy(data)
    if isinstance(data, (list, tuple)):
        return torch.as_tensor(data)
    logger.error(f"Unsupported tensor content type {type(data)} from {data_path}")
    raise TypeError("Unsupported tensor content type.")


def compare_tensor(golden_data_fp32, my_data_fp32):
    row_data: Dict[str, Any] = {}
    fail_messages: List[str] = []

    tensor_pass, message = check_tensor(golden_data_fp32, my_data_fp32)
    if not tensor_pass:
        logger.debug(f"check_tensor failed: {message}")
        row_data[CMP_FAIL_REASON] = message
        return row_data

    cmp_alg_map, custom_alg_map = _lazy_cmp_alg_maps()
    for name, cmp_func in list(cmp_alg_map.items()) + list(custom_alg_map.items()):
        result, message = cmp_func(golden_data_fp32, my_data_fp32)
        row_data[name] = result
        if message:
            fail_messages.append(message)
    row_data[CMP_FAIL_REASON] = " ".join(fail_messages)
    return row_data


def check_tensor(golden_data_fp32, my_data_fp32):
    torch = _lazy_import_torch()
    tensor_pass = True
    fail_reasons = []

    if len(golden_data_fp32) != len(my_data_fp32):
        fail_reasons.append("data shape doesn't match.")
        tensor_pass = False
    if not torch.all(torch.isfinite(golden_data_fp32)):
        fail_reasons.append("golden_data includes NAN or inf.")
        tensor_pass = False
    if not torch.all(torch.isfinite(my_data_fp32)):
        fail_reasons.append("my_data includes NAN or inf.")
        tensor_pass = False
    return tensor_pass, " ".join(fail_reasons)


__all__ = [
    "BasicDataInfo",
    "fill_row_data",
    "save_compare_result_to_csv",
]



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from abc import ABC, abstractmethod
import re

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import ScopeException


class ScopeFactory:
    def __init__(self, config):
        self.task = config.task
        self.level = config.level
        self.scope = config.scope
        self.api_list = config.list

    def build_scope(self):
        if not self.scope and not self.api_list:
            return None
        if self.scope is None:
            self.scope = []
        if self.api_list is None:
            self.api_list = []
        return self._build_range_scope()

    def _build_range_scope(self):
        api_range_scope = APIRangeScope(self.scope, self.api_list, self.level)
        module_range_scope = ModuleRangeScope(self.scope, self.api_list, self.level)
        mix_range_scope = MixRangeScope(self.scope, self.api_list, self.level)

        if self.level == Const.LEVEL_MIX:
            return mix_range_scope

        if not self.scope:
            return api_range_scope
        if api_range_scope.is_valid and module_range_scope.is_valid:
            raise ScopeException(ScopeException.InvalidScope, f"scope={self.scope}.")
        elif api_range_scope.is_valid:
            return api_range_scope
        elif module_range_scope.is_valid:
            return module_range_scope
        else:
            raise ScopeException(ScopeException.InvalidScope, f"scope={self.scope}")


class BaseScope(ABC):
    Module_Type_Module = "Module"
    Module_Type_API = "api"
    module_type = ["Module", "Cell"]

    def __init__(self, scope, api_list, level=None):
        self.level = level
        scope, api_list = self.rectify_args(scope, api_list)
        self.scope = scope
        self.api_list = api_list

    def rectify_args(self, scope, api_list):
        if not isinstance(api_list, list):
            raise ScopeException(ScopeException.InvalidApiStr,
                                 f"api_list参数须配置为列表，实际类型为{type(api_list)}.")
        for api in api_list:
            if not isinstance(api, str):
                raise ScopeException(ScopeException.InvalidApiStr,
                                     f"api_list中的元素须配置为字符串，实际类型为{type(api)}.")
        if isinstance(scope, str):
            scope = [scope]
            return scope, api_list
        if not isinstance(scope, list):
            raise ScopeException(ScopeException.InvalidScope,
                                 f"scope参数须配置为字符串或列表，实际类型为{type(scope)}.")
        for s in scope:
            if not isinstance(s, str):
                raise ScopeException(ScopeException.InvalidScope,
                                     f"scope列表元素要求类型为字符串，实际类型为{type(s)}.")
        return scope, api_list

    @abstractmethod
    def check(self, name):
        pass

    def check_api_list(self, api_name):
        if not self.api_list:
            return True
        for api_str in self.api_list:
            if api_str in api_name:
                return True
        return False


class ListScope(BaseScope):
    def rectify_args(self, scope, api_list):
        if scope and api_list:
            raise ScopeException(ScopeException.ArgConflict,
                                 f"scope和api_list不可以同时配置，实际配置为scope={scope}, api_list={api_list}.")
        return super().rectify_args(scope, api_list)

    def check(self, name):
        if not self.scope or name in self.scope:
            return self.check_api_list(name)
        return False


class RangeScope(BaseScope, ABC):

    def __init__(self, *args):
        super().__init__(*args)
        self.in_scope = False
        self.in_list = False
        self.start_name_set = set()
        self.is_valid = self.check_scope_is_valid()

    def check_name_pattern(self, name):
        options_pattern = "|".join(re.escape(option) for option in Const.DUMP_PREFIX)
        api_pattern = rf"^({options_pattern})\..*\.\d+\.(forward|backward)$"
        module_pattern = r"^(Cell|Module)\..*\.(forward|backward)\.\d+$"

        if self.level == Const.LEVEL_L1:
            if not re.match(api_pattern, name):
                raise ScopeException(ScopeException.InvalidScope,
                                     f"scope参数格式错误，要求格式为api完整命名，实际为{name}.")

        if self.level == Const.LEVEL_L0:
            if not re.match(module_pattern, name):
                raise ScopeException(ScopeException.InvalidScope,
                                     f"scope参数格式错误，要求格式为模块完整命名，实际为{name}.")

        if self.level == Const.LEVEL_MIX:
            if not re.match(api_pattern, name) and not re.match(module_pattern, name):
                raise ScopeException(ScopeException.InvalidScope,
                                     f"scope参数格式错误，要求格式为api或模块完整命名，实际为{name}.")

    def rectify_args(self, scope, api_list):
        scope, api_list = super().rectify_args(scope, api_list)
        if scope and len(scope) != 2:
            raise ScopeException(ScopeException.InvalidScope,
                                 f"scope参数指定区间断点，须传入长度为2的列表，实际长度为{len(scope)}.")
        for name in scope:
            self.check_name_pattern(name)
        return scope, api_list

    @abstractmethod
    def check_scope_is_valid(self):
        pass

    def begin_module(self, module_name):
        pass

    def end_module(self, module_name):
        pass


class APIRangeScope(RangeScope):
    def check_scope_is_valid(self):
        if not self.scope:
            return True
        scope_start_type = self.scope[0].split(Const.SEP)[0]
        if scope_start_type in BaseScope.module_type:
            return False
        scope_stop_type = self.scope[1].split(Const.SEP)[0]
        if scope_stop_type in BaseScope.module_type:
            return False
        return True

    def check(self, name):
        if self.scope and name == self.scope[0]:
            self.in_scope = True

        if not self.scope or self.in_scope:
            result = self.check_api_list(name)
        else:
            result = False

        if self.scope and name == self.scope[1]:
            self.in_scope = False
        return result


class ModuleRangeScope(RangeScope):
    """
        模块与api不同的是，模块内部还有子结构需要dump，
        需要用pre_hook和full_backward_hook来精确控制module的开始和结束，
        在这些hook触发时调用begin_module和end_module做区间控制
    """

    def check_scope_is_valid(self):
        if not self.scope:
            return True
        scope_start_type = self.scope[0].split(Const.SEP)[0]
        scope_stop_type = self.scope[1].split(Const.SEP)[0]
        if scope_start_type in BaseScope.module_type and \
                scope_stop_type in BaseScope.module_type:
            return True
        return False

    def begin_module(self, module_name):
        if not self.scope:
            return
        if module_name == self.scope[0]:
            self.in_scope = True

    def end_module(self, module_name):
        if not self.scope:
            return
        if module_name == self.scope[1]:
            self.in_scope = False

    def check(self, name):
        if not self.scope or self.in_scope:
            return self.check_api_list(name)
        return False


class MixRangeScope(RangeScope):
    def check_scope_is_valid(self):
        return True if self.scope else False

    def begin_module(self, module_name):
        if self.scope and module_name == self.scope[0]:
            self.in_scope = True
        for name in self.api_list:
            if name in module_name:
                self.in_list = True
                self.start_name_set.add(module_name)  # 记录每一个开启in_list的module_name

    def end_module(self, module_name):
        if self.scope and module_name == self.scope[1]:
            self.in_scope = False
        self.start_name_set.discard(module_name)  # 从集合中删除每一个module_name
        if not self.start_name_set:  # 如果集合为空，说明当前module_name是最后一个开启in_list的module_name
            self.in_list = False  # 关闭in_list

    def check_api_list(self, api_name):
        if not self.api_list:
            return True

        for name in self.api_list:
            if name in api_name:
                return True
        return False

    def check(self, name):
        """
        dump时调用的接口，根据scope和api_list判断是否需要dump
        """
        result = False
        if self.scope and name == self.scope[0]:
            self.in_scope = True

        if not self.scope or self.in_scope:
            if self.in_list:
                result = True
            else:
                result = self.check_api_list(name)

        if self.scope and name == self.scope[1]:
            self.in_scope = False
        return result



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
FusionRuleParser class.
This class mainly involves the analysis_fusion_rule function.
"""

import uuid

from msprobe.msaccucmp.cmp_utils import utils, utils_type
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.vector_cmp.fusion_manager.quant_filter import QuantFilter
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import OpAttr, FusionOp, OutputDesc
from msprobe.msaccucmp.dump_parse import dump


def make_left_and_right_string(ground_truth_to_my_output_map: dict) -> (str, str):
    """
    Convert my output ops and ground truth ops to string, split by ","
    :param ground_truth_to_my_output_map: the map for ground truth op to my output op list
    :return my output op names, ground truth op names
    """
    ground_truth_ops_str = ",".join(ground_truth_to_my_output_map.keys())
    my_output_op_name_list = []
    for op_list in ground_truth_to_my_output_map.values():
        for fusion_op in op_list:
            my_output_op_name_list.append(fusion_op.op_name)
    my_output_ops_str = ",".join(my_output_op_name_list)
    return my_output_ops_str, ground_truth_ops_str


def get_relation_for_fusion(fusion_op_list: list) -> utils_type.FusionRelation:
    """
    Get the relation for fusion op list
    :param fusion_op_list: the fusion op list
    :return FusionRelation
    """
    if len(fusion_op_list) == 1:
        if len(fusion_op_list[0].attr.original_op_names) > 1:
            return utils_type.FusionRelation.MultiToOne
        return utils_type.FusionRelation.OneToOne
    if len(fusion_op_list) > 1:
        if fusion_op_list[0].attr.l1_fusion_no != "":
            return utils_type.FusionRelation.L1Fusion
        if len(fusion_op_list[0].attr.original_op_names) == 1:
            return utils_type.FusionRelation.OneToMulti
        return utils_type.FusionRelation.MultiToMulti
    return utils_type.FusionRelation.OneToOne


def make_right_to_left_multi_map(fusion_op_list: list) -> dict:
    """
    make map {original_op_name_string, op_list} for any to multi
    :param fusion_op_list: the fusion op list
    :return the map {original_op_name_string, op_list}
    """
    # if multi to multi, the map is{'C,D,F': ['A', 'B']}
    # if one to multi, {'G': ['H', 'I']}
    # if splitter, the value of return is like {A: [A_1, A_2], B: [B_1, B_2]}
    ground_truth_to_my_output_map = {}
    for operator in fusion_op_list:
        original_names = ','.join(operator.attr.original_op_names)
        if original_names in ground_truth_to_my_output_map:
            ground_truth_to_my_output_map.get(original_names).append(operator)
        else:
            ground_truth_to_my_output_map[original_names] = [operator]
    return ground_truth_to_my_output_map


def _get_original_op_names_by_op_list(op_list: list, tmp_original_op_names: list) -> None:
    for fusion_op in op_list:
        for original_op_name in fusion_op.attr.original_op_names:
            if original_op_name not in tmp_original_op_names:
                tmp_original_op_names.append(original_op_name)


def _get_original_op_names_before_quant(fusion_op_info: FusionOp, quant_fusion_rule: any) -> list:
    tmp_original_op_names = []
    for op_name in fusion_op_info.attr.original_op_names:
        if op_name in quant_fusion_rule.op_name_to_fusion_op_name_map:
            op_list = quant_fusion_rule.fusion_op_name_to_op_map.get(
                quant_fusion_rule.op_name_to_fusion_op_name_map.get(op_name))
            _get_original_op_names_by_op_list(op_list, tmp_original_op_names)
    return tmp_original_op_names


def _make_new_output_desc(desc: OutputDesc, origin_format: str, origin_shape: list) -> None:
    if origin_format != '':
        desc.origin_format = origin_format
        desc.origin_shape = origin_shape


def _make_output_desc_by_index(output_desc: OutputDesc, op_list: list, tmp_output_desc: list) -> None:
    for operator in op_list:
        if operator.op_name != output_desc.origin_name:
            continue
        if output_desc.origin_output_index >= len(operator.output_desc):
            tmp_output_desc.append(OutputDesc("", None, "", []))
        else:
            desc = operator.output_desc[output_desc.origin_output_index]
            _make_new_output_desc(desc, output_desc.origin_format, output_desc.origin_shape)
            tmp_output_desc.append(desc)


def _make_output_desc_by_op_list(output_desc: OutputDesc, op_list: list, tmp_output_desc: list) -> None:
    origin_output_index = output_desc.origin_output_index
    # get origin output desc
    if origin_output_index is None:
        for item in op_list:
            for desc in item.output_desc:
                _make_new_output_desc(desc, output_desc.origin_format, output_desc.origin_shape)
                tmp_output_desc.append(desc)
    else:
        _make_output_desc_by_index(output_desc, op_list, tmp_output_desc)


def _get_output_desc_before_quant(fusion_op_info: FusionOp, quant_fusion_rule: any) -> list:
    tmp_output_desc = []
    for output_desc in fusion_op_info.output_desc:
        origin_name = output_desc.origin_name
        # skip empty origin name
        if origin_name == "":
            continue
        # skip origin name not in quant
        if origin_name not in quant_fusion_rule.op_name_to_fusion_op_name_map:
            log.print_warn_log('The name "%s" is not in quant fusion rule.' % origin_name)
            continue
        op_list = quant_fusion_rule.fusion_op_name_to_op_map.get(
            quant_fusion_rule.op_name_to_fusion_op_name_map.get(origin_name))
        _make_output_desc_by_op_list(output_desc, op_list, tmp_output_desc)
    return tmp_output_desc


def merge_fusion_rule(offline_fusion_rule: any, quant_fusion_rule: any) -> any:
    """
    Merge offline fusion rule and quant fusion rule
    :param offline_fusion_rule: the offline fusion rule
    :param quant_fusion_rule: the quant fusion rule
    :return the merged fusion rule
    """
    merge_map = {}
    for (_, fusion_op_list) in list(
            offline_fusion_rule.fusion_op_name_to_op_map.items()):
        for fusion_op in fusion_op_list:
            # get the quant fusion rule original op names, and deduplication
            tmp_original_op_names = _get_original_op_names_before_quant(
                fusion_op, quant_fusion_rule)
            # replace original op names
            if len(tmp_original_op_names) > 0:
                fusion_op.attr.original_op_names = list(tmp_original_op_names)

            # replace output desc
            tmp_output_desc = _get_output_desc_before_quant(fusion_op, quant_fusion_rule)
            if len(tmp_output_desc) > 0:
                fusion_op.output_desc = tmp_output_desc

            # replace fusion op name by original_op_names
            offline_fusion_rule.make_fusion_op_name(fusion_op.op_name, "", fusion_op.attr.original_op_names)
            fusion_op_name = offline_fusion_rule.op_name_to_fusion_op_name_map.get(fusion_op.op_name)
            if fusion_op_name in merge_map:
                fusion_op.op_id = merge_map.get(fusion_op_name)[0].op_id
                merge_map.get(fusion_op_name).append(fusion_op)
            else:
                fusion_op.op_id = len(merge_map)
                merge_map[fusion_op_name] = [fusion_op]
    offline_fusion_rule.fusion_op_name_to_op_map = merge_map
    return offline_fusion_rule


def _make_open_fusion_original_op_names(fusion_op: FusionOp, close_fusion_origin_to_op_map: dict) -> list:
    if len(fusion_op.attr.original_op_names) == 1 and fusion_op.attr.original_op_names[0] == '':
        return ['']
    original_op_names = []
    for origin_op_name in fusion_op.attr.original_op_names:
        if origin_op_name in close_fusion_origin_to_op_map:
            op_name = close_fusion_origin_to_op_map.get(origin_op_name)
            if op_name not in original_op_names:
                original_op_names.append(op_name)
        else:
            log.print_warn_log('There is no original operator associated with the operator "%s" '
                               'in original op names.' % origin_op_name)
    return original_op_names


def _get_close_fusion_origin_output_index(op_list: list, output_desc: OutputDesc) -> bool:
    for close_op in op_list:
        for index, close_output_desc in enumerate(close_op.output_desc):
            if output_desc.origin_name == close_output_desc.origin_name \
                    and output_desc.origin_output_index == close_output_desc.origin_output_index:
                output_desc.origin_output_index = index
                return True
    return False


def _make_open_fusion_output_desc(fusion_op: FusionOp, close_fusion_origin_to_op_map: dict,
                                  close_fusion_rule: any) -> None:
    for output_index, output_desc in enumerate(fusion_op.output_desc):
        if not output_desc.origin_name:
            continue
        if output_desc.origin_name in close_fusion_origin_to_op_map:
            op_name = close_fusion_origin_to_op_map.get(output_desc.origin_name)
            op_list = close_fusion_rule.fusion_op_name_to_op_map.get(
                close_fusion_rule.op_name_to_fusion_op_name_map.get(op_name))
            if not _get_close_fusion_origin_output_index(op_list, output_desc):
                log.print_warn_log('There is no valid output desc associated with "%s:output:%d".'
                                   % (fusion_op.op_name, output_index))
            output_desc.origin_name = op_name
        else:
            log.print_warn_log('There is no valid output desc associated with "%s:output:%d".'
                               % (fusion_op.op_name, output_index))


def _check_unity_onnx_op(delete_open_fusion_op_name_to_op_name: list, delete_close_fusion_op_name_to_op_name: list,
                         fusion_op: FusionOp, close_fusion_origin_to_op_map: list) -> bool:
    if fusion_op.attr.original_op_names[0] in delete_open_fusion_op_name_to_op_name:
        return False
    delete_open_fusion_op_name_to_op_name.append(fusion_op.attr.original_op_names[0])
    if fusion_op.attr.original_op_names[0] not in close_fusion_origin_to_op_map:
        return True
    if close_fusion_origin_to_op_map.get(fusion_op.attr.original_op_names[0]) in delete_close_fusion_op_name_to_op_name:
        return False
    delete_close_fusion_op_name_to_op_name.append(close_fusion_origin_to_op_map.get(
        fusion_op.attr.original_op_names[0]))
    return True




def merge_close_and_open_fusion_rule(open_fusion_rule: any, close_fusion_rule: any) -> any:
    """
    Merge close fusion rule and open fusion rule
    :param open_fusion_rule: the open fusion rule
    :param close_fusion_rule: the close fusion rule
    :return the merged fusion rule
    """
    merged_fusion_rule = FusionRuleParser('')
    delete_fusion_op_name_list = []
    delete_open_fusion_op_name_to_op_name = []
    delete_close_fusion_op_name_to_op_name = []
    close_fusion_origin_to_op_map = close_fusion_rule.get_origin_name_to_op_name_map()
    for key, value in reversed(list(open_fusion_rule.op_name_to_fusion_op_name_map.items())):
        if value in delete_fusion_op_name_list:
            continue
        delete_fusion_op_name_list.append(value)
        merged_fusion_rule.op_name_to_fusion_op_name_map[key] = value
        for fusion_op in open_fusion_rule.fusion_op_name_to_op_map.get(value):
            if not _check_unity_onnx_op(delete_open_fusion_op_name_to_op_name, delete_close_fusion_op_name_to_op_name,
                                        fusion_op, close_fusion_origin_to_op_map):
                continue
            # make new original op names
            fusion_op.attr.original_op_names = _make_open_fusion_original_op_names(fusion_op,
                                                                                   close_fusion_origin_to_op_map)
            # make new output desc
            _make_open_fusion_output_desc(fusion_op, close_fusion_origin_to_op_map, close_fusion_rule)
            if value in merged_fusion_rule.fusion_op_name_to_op_map:
                merged_fusion_rule.fusion_op_name_to_op_map.get(value).append(fusion_op)
            else:
                merged_fusion_rule.fusion_op_name_to_op_map[value] = [fusion_op]
            merged_fusion_rule.op_list.insert(0, fusion_op)
    return merged_fusion_rule


class FusionRuleParser:
    """
    the class for parse fusion rule.
    """

    def __init__(self: any, path: str) -> None:
        self.json_path = path
        self.json_object = None
        self.fusion_op_name_to_op_map = {}
        self.op_name_to_fusion_op_name_map = {}
        self.op_list = []
        self.input_nodes = []

    @staticmethod
    def _check_key_exist(json_object: any, key: str) -> None:
        if key not in json_object:
            log.print_warn_log('There is no "%s" element in fusion rule file.' % key)
            raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR)

    @staticmethod
    def _make_output_desc(output_desc_list: list, name: str) -> None:
        if len(output_desc_list) == 0:
            output_desc = OutputDesc(name, None, "", [])
            output_desc_list.append(output_desc)
        else:
            for (index, _) in enumerate(output_desc_list):
                if output_desc_list[index].origin_name == "":
                    output_desc_list[index].origin_name = name

    @staticmethod
    def _process_ffts_op_name(item):
        if ConstManager.FFTS_MANUAL_MODE_FIELD in item:
            item = dump.process_op_name(item)
        return item

    def analysis_fusion_rule(self: any) -> None:
        """
        Analysis fusion json file
        """
        self.json_object = FileUtils.load_json_file(self.json_path)
        self._parse_fusion_op_json_object()
        # analysis and filter the parsed op list
        filtering = QuantFilter(self.op_list)
        filtering.process_filtering()

    def make_fusion_op_name(self: any, name: str, l1_fusion_no: str, original_op_names: list) -> None:
        """
        Make fusion op name by group op name and original op names
        :return the fusion op name
        """
        # the fusion op name priority:
        # l1_fusion_no -> original_op_names -> name
        if l1_fusion_no != "":
            # the l1_fusion_no is not empty,
            # the fusion op name is the l1_fusion_no
            self.op_name_to_fusion_op_name_map[name] = l1_fusion_no
            return

        if original_op_names:
            if len(original_op_names) == 1:
                # There is one original op name
                if original_op_names[0] == '':
                    # the original name is empty, the fusion op name is op name
                    self.op_name_to_fusion_op_name_map[name] = name
                else:
                    # the original name is not empty,
                    # the fusion op name is original op name
                    self.op_name_to_fusion_op_name_map[name] = original_op_names[0]
            else:
                # The original op name more than one,
                # the fusion op name is uuid names
                self.op_name_to_fusion_op_name_map[name] = uuid.uuid3(uuid.NAMESPACE_DNS, ''.join(original_op_names))
        else:
            self.op_name_to_fusion_op_name_map[name] = name

    def get_origin_name_to_op_name_map(self: any) -> dict:
        """
        Get origin name to op name map
        :return: the map
        """
        origin_name_to_op_name_map = {}
        for fusion_op in self.op_list:
            for origin_name in fusion_op.attr.original_op_names:
                origin_name_to_op_name_map[origin_name] = fusion_op.op_name
        return origin_name_to_op_name_map

    def check_array_object_valid(self: any, json_object: any, key: str) -> None:
        """
        Check array object valid
        :param json_object:the json object
        :param key : key in json
        """
        self._check_key_exist(json_object, key)
        if not isinstance(json_object[key], list):
            log.print_error_log('The content of the json file "%r" is invalid. The "%s" element is not an array.'
                                % (self.json_path, key))
            raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR)

    def check_string_object_valid(self: any, json_object: any, key: str) -> None:
        """
        Check string object valid
        :param json_object:the json object
        :param key : key in json
        """
        self._check_key_exist(json_object, key)
        if not isinstance(json_object[key], str):
            log.print_error_log('The content of the json file "%r" is invalid. The "%s" element is not a string.'
                                % (self.json_path, key))
            raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR)

    def get_fusion_op_list(self: any, op_name: str) -> (list, FusionOp):
        """
        Get the fusion op list by op name
        :param op_name: the op name
        :return :the fusion op list, the fusion op by name
        """
        if op_name not in self.op_name_to_fusion_op_name_map:
            message = 'There is no "%s" in the fusion rule file.' % op_name
            log.print_warn_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, message)
        fusion_op_name = self.op_name_to_fusion_op_name_map.get(op_name)
        fusion_op_list = self.fusion_op_name_to_op_map.get(fusion_op_name, [])

        # get fusion op in list by op name
        fusion_op_info = None
        for operator in fusion_op_list:
            if operator.op_name == op_name:
                fusion_op_info = operator
                break
        if fusion_op_info is None:
            message = 'There is no "%s" in the fusion rule file.' % op_name
            log.print_warn_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR, message)
        return fusion_op_list, fusion_op_info

    def _adjust_rename_node(self: any) -> None:
        for _, fusion_op_list in self.fusion_op_name_to_op_map.items():
            if len(fusion_op_list) == 1 and self._is_rename_node(fusion_op_list[0]):
                self._make_output_desc(fusion_op_list[0].output_desc, fusion_op_list[0].attr.original_op_names[0])

    def _parse_fusion_op_json_object(self: any) -> None:
        # check graph element in json file
        self.check_array_object_valid(self.json_object, ConstManager.GRAPH_OBJECT)
        for graph in self.json_object[ConstManager.GRAPH_OBJECT]:
            # check op element in graph value
            self.check_array_object_valid(graph, ConstManager.OP_OBJECT)
            for operator in graph[ConstManager.OP_OBJECT]:
                self._parse_op_object(operator)

            # adjust the output desc for the rename node
            self._adjust_rename_node()

        self.op_list.sort(key=lambda x: x.attr.get_op_sequence())

    def _parse_input_nodes(self, op_object):
        if ConstManager.DATA_OBJECT == op_object.get(ConstManager.TYPE_OBJECT):
            if op_object.get(ConstManager.NAME_OBJECT):
                self.input_nodes.append(op_object.get(ConstManager.NAME_OBJECT))

    def _parse_input(self: any, op_object: any) -> list:
        input_list = []
        # data layer has no input layer
        if ConstManager.INPUT_OBJECT in op_object:
            if not isinstance(op_object[ConstManager.INPUT_OBJECT], list):
                log.print_error_log('The content of the json file "%r" is invalid. The "%s" element is not '
                                    'an array.' % (self.json_path, ConstManager.INPUT_OBJECT))
                raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR)
            for item in op_object[ConstManager.INPUT_OBJECT]:
                if item == "" and len(op_object[ConstManager.INPUT_OBJECT]) == 1:
                    break
                # skip control edge
                if not item.endswith(':-1') and item != "":
                    item = self._process_ffts_op_name(item)
                    input_list.append(item)
        return input_list

    def _parse_output_desc_in_attr(self: any, output_desc_attr: any, default_index: int) -> OutputDesc:
        origin_name = self._get_string_value_in_attr(output_desc_attr, ConstManager.ORIGIN_NAME_OBJECT)
        origin_output_index = self._get_int_value_in_attr(output_desc_attr, ConstManager.ORIGIN_OUTPUT_INDEX_OBJECT)
        if origin_output_index is None:
            origin_output_index = default_index
        origin_output_format = self._get_string_value_in_attr(output_desc_attr, ConstManager.ORIGIN_FORMAT_OBJECT)
        if origin_output_format == '':
            origin_output_format = self._get_string_value_in_attr(
                output_desc_attr, ConstManager.GE_ORIGIN_FORMAT_OBJECT)
        origin_output_shape = self._get_origin_shape_in_attr(output_desc_attr)
        return OutputDesc(origin_name, origin_output_index, origin_output_format, origin_output_shape)

    def _parse_output_desc(self: any, op_object: any) -> list:
        output_desc_list = []
        # get output desc
        if ConstManager.OUTPUT_DESC_OBJECT in op_object:
            default_index = 0
            for output_desc_object in op_object[ConstManager.OUTPUT_DESC_OBJECT]:

                d_type = ""
                if ConstManager.D_TYPE in output_desc_object:
                    d_type = output_desc_object.get(ConstManager.D_TYPE)

                if ConstManager.ATTR_OBJECT in output_desc_object:
                    output_desc = self._parse_output_desc_in_attr(
                        output_desc_object[ConstManager.ATTR_OBJECT], default_index)
                    output_desc.set_data_type(d_type)
                    output_desc_list.append(output_desc)
                default_index += 1
        return output_desc_list

    def _is_rename_node(self: any, fusion_op: FusionOp) -> bool:
        return len(fusion_op.attr.original_op_names) == 1 and \
               self.op_name_to_fusion_op_name_map.get(fusion_op.op_name) == fusion_op.attr.original_op_names[0]

    def _parse_attr(self: any, op_object: any, op_name: str) -> (OpAttr, bool):
        # check attr element is valid
        if ConstManager.ATTR_OBJECT not in op_object:
            attr_array = []
        else:
            self.check_array_object_valid(op_object, ConstManager.ATTR_OBJECT)
            attr_array = op_object[ConstManager.ATTR_OBJECT]
        is_multi_op = self._get_bool_value_in_attr(attr_array, ConstManager.IS_MULTI_OP)
        # get l1_fusion_sub_graph_no
        l1_fusion_no = self._get_string_value_in_attr(attr_array, ConstManager.L1_FUSION_SUB_GRAPH_NO_OBJECT)
        # get original op names
        original_op_names, have_origin = self._get_original_op_names_in_attr(attr_array, op_name)
        op_sequence = self._parse_id_object(op_object)
        return OpAttr(original_op_names, l1_fusion_no, is_multi_op, op_sequence), have_origin

    def _parse_id_object(self: any, op_object: any) -> int:
        op_sequence = 0
        if ConstManager.ID_OBJECT in op_object:
            self._check_int_object_valid(op_object, ConstManager.ID_OBJECT)
            op_sequence = op_object[ConstManager.ID_OBJECT]
        return op_sequence

    def _parse_op_object(self: any, op_object: dict) -> None:
        # check name element is valid
        self.check_string_object_valid(op_object, ConstManager.NAME_OBJECT)
        name = op_object[ConstManager.NAME_OBJECT]
        if ConstManager.FFTS_MANUAL_MODE_FIELD in name:
            name = dump.process_op_name(name)
        # check type element is valid
        self.check_string_object_valid(op_object, ConstManager.TYPE_OBJECT)
        self._parse_input_nodes(op_object)

        input_list = self._parse_input(op_object)

        output_desc_list = self._parse_output_desc(op_object)
        attr, have_origin = self._parse_attr(op_object, name)
        if not have_origin:
            self._make_output_desc(output_desc_list, name)

        self.make_fusion_op_name(name, attr.l1_fusion_no, attr.original_op_names)
        fusion_op_name = self.op_name_to_fusion_op_name_map.get(name)
        fusion_op = FusionOp(0, name, input_list, op_object[ConstManager.TYPE_OBJECT], output_desc_list, attr)
        if fusion_op_name in self.fusion_op_name_to_op_map:
            fusion_op.op_id = self.fusion_op_name_to_op_map.get(fusion_op_name)[0].op_id
            fusion_op_name_list = \
                [_fusion_op.op_name for _fusion_op in self.fusion_op_name_to_op_map.get(fusion_op_name)]
            if name not in fusion_op_name_list:
                self.fusion_op_name_to_op_map.get(fusion_op_name).append(fusion_op)
                self.op_list.append(fusion_op)
        else:
            fusion_op.op_id = len(self.fusion_op_name_to_op_map)
            self.fusion_op_name_to_op_map[fusion_op_name] = [fusion_op]
            self.op_list.append(fusion_op)

    def _get_string_value_in_attr(self: any, attr_array: list, key: str) -> str:
        value = ""
        for attr in attr_array:
            self.check_string_object_valid(attr, ConstManager.KEY_OBJECT)
            key_value = attr[ConstManager.KEY_OBJECT]
            if key_value == key:
                self._check_key_exist(attr, ConstManager.VALUE_OBJECT)
                value_value = attr[ConstManager.VALUE_OBJECT]
                self.check_string_object_valid(value_value, ConstManager.STRING_TYPE_OBJECT)
                value = value_value[ConstManager.STRING_TYPE_OBJECT]
                break
        return value

    def _get_int_value_in_attr(self: any, attr_array: list, key: str) -> int:
        value = None
        for attr in attr_array:
            self.check_string_object_valid(attr, ConstManager.KEY_OBJECT)
            key_value = attr[ConstManager.KEY_OBJECT]
            if key_value == key:
                self._check_key_exist(attr, ConstManager.VALUE_OBJECT)
                value_value = attr[ConstManager.VALUE_OBJECT]
                self._check_int_object_valid(value_value, ConstManager.INT_TYPE_OBJECT)
                value = value_value[ConstManager.INT_TYPE_OBJECT]
                break
        return value

    def _get_origin_shape_in_attr(self: any, attr_array: list) -> list:
        value = []
        for attr in attr_array:
            self.check_string_object_valid(attr, ConstManager.KEY_OBJECT)
            key_value = attr[ConstManager.KEY_OBJECT]
            if key_value == ConstManager.GE_ORIGIN_SHAPE_OBJECT:
                self._check_key_exist(attr, ConstManager.VALUE_OBJECT)
                value_value = attr[ConstManager.VALUE_OBJECT]
                self._check_key_exist(value_value, ConstManager.LIST_TYPE_OBJECT)
                if ConstManager.INT_TYPE_OBJECT in value_value[ConstManager.LIST_TYPE_OBJECT]:
                    self.check_array_object_valid(
                        value_value[ConstManager.LIST_TYPE_OBJECT], ConstManager.INT_TYPE_OBJECT)
                    value = value_value[ConstManager.LIST_TYPE_OBJECT][ConstManager.INT_TYPE_OBJECT]
                break
        return value

    def _get_bool_value_in_attr(self: any, attr_array: list, key: str) -> bool:
        value = False
        for attr in attr_array:
            self.check_string_object_valid(attr, ConstManager.KEY_OBJECT)
            key_value = attr[ConstManager.KEY_OBJECT]
            if key_value == key:
                self._check_key_exist(attr, ConstManager.VALUE_OBJECT)
                value_value = attr[ConstManager.VALUE_OBJECT]
                self._check_bool_object_valid(value_value, ConstManager.BOOL_TYPE_OBJECT)
                value = value_value[ConstManager.BOOL_TYPE_OBJECT]
                break
        return value

    def _get_original_op_names_in_attr(self: any, attr_array: list, op_name: str) -> (list, bool):
        array = []
        match = False
        for attr in attr_array:
            self.check_string_object_valid(attr, ConstManager.KEY_OBJECT)
            key_value = attr[ConstManager.KEY_OBJECT]
            if key_value == ConstManager.ORIGINAL_OP_NAMES_OBJECT:
                self._check_key_exist(attr, ConstManager.VALUE_OBJECT)
                value = attr[ConstManager.VALUE_OBJECT]
                self._check_key_exist(value, ConstManager.LIST_TYPE_OBJECT)
                if ConstManager.STRING_TYPE_OBJECT not in value[ConstManager.LIST_TYPE_OBJECT]:
                    array = ['']
                else:
                    self.check_array_object_valid(value[ConstManager.LIST_TYPE_OBJECT], ConstManager.STRING_TYPE_OBJECT)
                    array = value[ConstManager.LIST_TYPE_OBJECT][ConstManager.STRING_TYPE_OBJECT]
                match = True
                break
        if not match:
            array.append(op_name)
        return array, match

    def _check_int_object_valid(self: any, json_object: any, key: str) -> None:
        self._check_key_exist(json_object, key)
        if not isinstance(json_object[key], int):
            log.print_error_log('The content of the json file "%r" is invalid. The "%s" element is not a integer.'
                                % (self.json_path, key))
            raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR)

    def _check_bool_object_valid(self: any, json_object: any, key: str) -> None:
        self._check_key_exist(json_object, key)
        if not isinstance(json_object[key], bool):
            log.print_error_log('The content of the json file "%r" is invalid. The "%s" element is not a bool.'
                                % (self.json_path, key))
            raise CompareError(CompareError.MSACCUCMP_PARSER_JSON_FILE_ERROR)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import inspect

from mindspore import Tensor, ops, mint
from mindspore.mint import distributed
from mindspore.mint.nn import functional
from mindspore.communication import comm_func

from msprobe.core.common.file_utils import load_yaml
from msprobe.core.common.utils import Const
from msprobe.core.dump.api_dump.api_registry import ApiRegistry
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.common.const import Const as MsConst
from msprobe.mindspore.common.utils import is_mindtorch
from msprobe.mindspore.dump.dump_processor.hook_cell.hook_cell import HOOKCell


stub_tensor_existed = True
try:
    from mindspore.common._stub_tensor import StubTensor
except ImportError:
    stub_tensor_existed = False

cur_path = os.path.dirname(os.path.realpath(__file__))
if not is_mindtorch():
    _api_types = {
        Const.MS_FRAMEWORK: {
            Const.MS_API_TYPE_OPS: ((ops,), (ops,)),
            Const.MS_API_TYPE_TENSOR: ((Tensor,), (Tensor,)),
            Const.MS_API_TYPE_MINT: ((mint,), (mint,)),
            Const.MS_API_TYPE_MINT_FUNC: ((functional,), (functional,)),
            Const.MS_API_TYPE_COM: ((comm_func,), (comm_func,)),
            Const.MS_API_TYPE_MINT_DIST: ((distributed,), (distributed,))
        }
    }
    if stub_tensor_existed:
        _api_types.get(Const.MS_FRAMEWORK).update(
            {Const.MS_API_TYPE_STUB_TENSOR: ((StubTensor,), (StubTensor,))}
        )

    _supported_api_list_path = (os.path.join(cur_path, MsConst.SUPPORTED_API_LIST_FILE),)
    _blacklist = []
else:
    import torch
    import torch_npu
    _api_types = {
        Const.MT_FRAMEWORK: {
            Const.PT_API_TYPE_FUNCTIONAL: ((torch.nn.functional,), (torch.nn.functional,)),
            Const.PT_API_TYPE_TENSOR: ((torch.Tensor,), (torch.Tensor,)),
            Const.PT_API_TYPE_TORCH: ((torch,), (torch,)),
            Const.PT_API_TYPE_NPU: ((torch_npu,), (torch_npu,)),
            Const.PT_API_TYPE_DIST: ((torch.distributed,), (torch.distributed, torch.distributed.distributed_c10d))
        }
    }
    _supported_api_list_path = (os.path.join(cur_path, '../../../../pytorch/dump/api_dump',
                                             MsConst.SUPPORTED_API_LIST_FILE),)
    _blacklist = []

_inner_used_api = {
    Const.MS_FRAMEWORK + Const.SEP + Const.MS_API_TYPE_OPS: (
        ops, "norm", "square", "sqrt", "is_complex", "stack", "is_floating_point"
    ),
    Const.MS_FRAMEWORK + Const.SEP + Const.MS_API_TYPE_TENSOR: (
        Tensor, "to", "numel", 'sum'
    ),
    Const.MS_FRAMEWORK + Const.SEP + Const.MS_API_TYPE_MINT: (
        mint, "max", "min", "mean", "norm"
    )
}


class ApiTemplate(HOOKCell):
    def __init__(self, api_name, api_func, prefix, hook_build_func):
        self.api_name = api_name
        self.prefix_api_name = prefix + Const.SEP + str(api_name.split(Const.SEP)[-1]) + Const.SEP
        distributed_prefix = Const.DIST_API_TYPE_PREFIX if is_mindtorch() else Const.MINT_DIST_API_TYPE_PREFIX
        self.op_is_distributed = prefix == distributed_prefix
        super().__init__(hook_build_func)
        self.api_func = api_func

    @staticmethod
    def async_to_sync(output):
        # Fake handle, used to return after the CommHandle executes the wait method
        fake_handle = type("FakeHandle", (), {"wait": lambda self: None})()
        if isinstance(output, tuple) and len(output) == 2 and hasattr(output[1], "wait"):
            output[1].wait()
            output = (output[0], fake_handle)
        elif hasattr(output, "wait"):
            output.wait()
            output = fake_handle
        return output

    def construct(self, *args, **kwargs):
        if self.api_name.startswith(MsConst.DROPOUT_API_NAME_PREFIX):
            return args[0] if args else kwargs.get(Const.INPUT)

        output = self.api_func(*args, **kwargs)

        if self.prefix_api_name.startswith(
            (MsConst.DISTRIBUTED_DATA_PREFIX, Const.MINT_DIST_API_TYPE_PREFIX)
        ):
            try:
                bound = inspect.signature(self.api_func).bind(*args, **kwargs)
                bound.apply_defaults()
                use_async_op_flag = bound.arguments.get("async_op", False)
            except Exception as e:
                use_async_op_flag = False
                logger.warning(f"fail to get dist api's func signature because {e}, no wait")

            if use_async_op_flag or self.api_name in ["isend", "irecv"]:
                output = self.async_to_sync(output)
            if self.api_name == "batch_isend_irecv" and isinstance(output, list):
                output = [self.async_to_sync(handle) for handle in output]

        return output

    def forward(self, *args, **kwargs):
        if self.api_name.startswith(MsConst.DROPOUT_API_NAME_PREFIX):
            return args[0] if args else kwargs.get(Const.INPUT)
        return self.api_func(*args, **kwargs)


api_register = None
stub_tensor_set = False


def get_api_register(return_new=False):
    global stub_tensor_set

    def stub_method(method):
        def wrapped_method(*args, **kwargs):
            return method(*args, **kwargs)
        return wrapped_method
    if not is_mindtorch() and stub_tensor_existed and not stub_tensor_set:
        api_names = load_yaml(_supported_api_list_path[0]).get(Const.MS_API_TYPE_TENSOR, [])
        for attr_name in dir(StubTensor):
            attr = getattr(StubTensor, attr_name)
            if attr_name in api_names and callable(attr):
                setattr(StubTensor, attr_name, stub_method(attr))
        stub_tensor_set = True

    if return_new:
        return ApiRegistry(
            _api_types,
            _inner_used_api,
            _supported_api_list_path,
            ApiTemplate,
            _blacklist
        )

    global api_register
    if api_register is None:
        api_register = ApiRegistry(
            _api_types,
            _inner_used_api,
            _supported_api_list_path,
            ApiTemplate,
            _blacklist
        )
    return api_register



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import numpy as np


class CompareInput:
    """
    A class to encapsulate the input data required for comparison operations.

    Attributes:
        bench_output (np.ndarray): The benchmark output values.
        device_output (np.ndarray): The device output values.
        compare_column (class): A clasee to store and update comparison metrics.
        dtype (type, optional): The data type of the outputs. Defaults to None.
        rel_err_orign (float or array-like, optional): The original relative error values. Defaults to None.

    Methods:
        __init__(bench_output, device_output, compare_column, dtype, rel_err_orign): 
        Initializes an instance of CompareInput.
    """
    def __init__(self, bench_output, device_output, compare_column, dtype=None, rel_err_orign=None):
        self.bench_output = bench_output
        self.device_output = device_output
        if not isinstance(bench_output, np.ndarray) or not isinstance(device_output, np.ndarray):
            raise TypeError("The input should be numpy array")
        self.compare_column = compare_column
        self.dtype = dtype
        self.rel_err_orign = rel_err_orign


class PrecisionCompareInput:
    def __init__(self, row_npu, row_gpu, dtype, compare_column, api_name=None):
        self.row_npu = row_npu
        self.row_gpu = row_gpu
        self.dtype = dtype
        self.compare_column = compare_column
        self.api_name = api_name



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

# 标准库导入
import multiprocessing
from multiprocessing import Manager
import os
import signal
import sys
import time

# 第三方库导入
from mindspore import context
import numpy as np
from tqdm import tqdm

# 本地应用/库特定导入
from msprobe.core.common.const import Const, CompareConst
from msprobe.mindspore.api_accuracy_checker.api_accuracy_checker import ApiAccuracyChecker, BasicInfoAndStatus
from msprobe.mindspore.api_accuracy_checker.multi_data_manager import MultiDataManager
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.common.const import MsCompareConst

from msprobe.core.dump.data_dump.data_collector import build_data_collector
from msprobe.core.common.utils import Const, print_tools_ends_info, DumpPathAggregation


class MultiApiAccuracyChecker(ApiAccuracyChecker):
    def __init__(self, args):
        super().__init__(args)
        # 可以添加 MultiApiAccuracyChecker 特有的属性或方法
        self.api_infos = dict()

        # 使用 Manager 创建共享变量，确保进程间的同步
        self.manager = Manager()
        self.is_first_write = self.manager.Value('b', True)  # 创建共享变量

        # 初始化 DataManager 时传入共享的 is_first_write
        self.multi_data_manager = MultiDataManager(args.out_path, args.result_csv_path, self.is_first_write)

        self.args = args  # 将 args 保存为类的属性

        # 初始化一个属性来存储当前的设备ID（用于日志中显示）
        self.current_device_id = None

        self.save_error_data = args.save_error_data
        if self.save_error_data:
            config, dump_path_aggregation = self.init_save_error_data(args)
            self.data_collector = build_data_collector(config)
            self.data_collector.update_dump_paths(dump_path_aggregation)

    def process_on_device(self, device_id, api_infos, progress_queue):
        """
        在特定设备上处理一部分API。

        参数:
            device_id (int): 要使用的设备ID。
            api_infos (list): 包含API名称和对应信息的元组列表。
            progress_queue (multiprocessing.Queue): 用于通信进度更新的队列。
        """

        # 设置当前设备ID
        self.current_device_id = device_id

        # 设置 MindSpore context 的 device_id
        context.set_context(device_id=device_id)

        # 遍历当前进程分配的任务
        for _, (api_name_str, api_info) in enumerate(api_infos):
            logger.debug(f"Processing API: {api_name_str}, Device: {device_id}")

            if not self.multi_data_manager.is_unique_api(api_name_str):
                logger.debug(f"API {api_name_str} is not unique, skipping.")
                progress_queue.put(1)
                continue

            # 处理前向
            forward_output_list = self.process_forward(api_name_str, api_info)
            if forward_output_list is not Const.EXCEPTION_NONE:
                self.multi_data_manager.record(forward_output_list)

            # 处理反向
            backward_output_list = self.process_backward(api_name_str, api_info)
            if backward_output_list is not Const.EXCEPTION_NONE:
                self.multi_data_manager.record(backward_output_list)

            # 保存结果
            self.multi_data_manager.save_results(api_name_str)
            progress_queue.put(1)  # 更新进度

    def run_and_compare(self):
        # 获取要使用的设备ID列表
        device_ids = self.args.device_id

        # 按设备数划分要处理的 API 项
        partitioned_api_infos = list(self.api_infos.items())

        # 在主进程中进行交叉任务切分（基于取模的方式）
        partitioned_api_infos_split = [[] for _ in range(len(device_ids))]
        for idx, api_info in enumerate(partitioned_api_infos):
            device_index = idx % len(device_ids)  # 使用取模方法分配任务
            partitioned_api_infos_split[device_index].append(api_info)

        # 创建一个共享进度队列
        progress_queue = multiprocessing.Queue()

        # 进度条
        total_tasks = len(partitioned_api_infos)  # 计算总任务数
        with tqdm(total=total_tasks, desc="Total Progress", ncols=100) as pbar:
            # 创建多进程
            processes = []
            for index, device_id in enumerate(device_ids):
                process = multiprocessing.Process(target=self.process_on_device,
                                                  args=(device_id, partitioned_api_infos_split[index], progress_queue))
                processes.append(process)
                process.start()

            # 主进程更新进度条
            completed_tasks = 0
            while completed_tasks < total_tasks:
                try:
                    completed_tasks += progress_queue.get(timeout=Const.PROGRESS_TIMEOUT)  # 设置超时时间（秒）
                    pbar.update(1)
                except multiprocessing.queues.Empty:
                    logger.error("Timeout while waiting for progress updates. Skipping remaining tasks.")
                    break

                # 检查子进程状态
                for process in processes:
                    if not process.is_alive():
                        if process.exitcode != 0:
                            logger.error(f"Process {process.pid} exited with code {process.exitcode}.")
                            total_tasks -= len(partitioned_api_infos_split[processes.index(process)])
                        processes.remove(process)

            # 确保所有子进程完成或终止
            for process in processes:
                process.join(timeout=Const.PROGRESS_TIMEOUT)
                if process.is_alive():
                    logger.error(f"Process {process.pid} did not terminate. Forcing termination.")
                    process.terminate()

    def process_forward(self, api_name_str, api_info):
        """
        Overrides the parent class's process_forward method to log the device ID when exceptions occur.

        Parameters:
            api_name_str (str): The name of the API.
            api_info (object): The API information object.

        Returns:
            list or None: The forward output list or None if an error occurs.
        """
        if not api_info.check_forward_info():
            logger.debug(
                f"[Device {self.current_device_id}] API: {api_name_str} lacks forward information, skipping "
                f"forward check.")
            return Const.EXCEPTION_NONE

        try:
            forward_inputs_aggregation = self.prepare_api_input_aggregation(api_info, Const.FORWARD)
        except Exception as e:
            logger.warning(
                f"[Device {self.current_device_id}] Exception occurred while getting forward API inputs for "
                f"{api_name_str}. Skipping forward check. Detailed exception information: {e}.")
            return Const.EXCEPTION_NONE

        forward_output_list = None
        try:
            forward_output_list = self.run_and_compare_helper(api_info, api_name_str, forward_inputs_aggregation,
                                                              Const.FORWARD)
        except Exception as e:
            logger.warning(
                f"[Device {self.current_device_id}] Exception occurred while running and comparing {api_name_str} "
                f"forward API. Detailed exception information: {e}.")
        return forward_output_list

    def process_backward(self, api_name_str, api_info):
        """
        Overrides the parent class's process_backward method to log the device ID when exceptions occur.

        Parameters:
            api_name_str (str): The name of the API.
            api_info (object): The API information object.

        Returns:
            list or None: The backward output list or None if an error occurs.
        """
        if not api_info.check_backward_info():
            logger.debug(
                f"[Device {self.current_device_id}] API: {api_name_str} lacks backward information, skipping "
                f"backward check.")
            return Const.EXCEPTION_NONE

        try:
            backward_inputs_aggregation = self.prepare_api_input_aggregation(api_info, Const.BACKWARD)
        except Exception as e:
            logger.warning(
                f"[Device {self.current_device_id}] Exception occurred while getting backward API inputs for "
                f"{api_name_str}. Skipping backward check. Detailed exception information: {e}.")
            return Const.EXCEPTION_NONE

        backward_output_list = None
        try:
            backward_output_list = self.run_and_compare_helper(api_info, api_name_str, backward_inputs_aggregation,
                                                               Const.BACKWARD)
        except Exception as e:
            logger.warning(
                f"[Device {self.current_device_id}] Exception occurred while running and comparing {api_name_str} "
                f"backward API. Detailed exception information: {e}.")
        return backward_output_list